# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


import equinox as eqx
import jax.numpy as jnp
from jax.random import PRNGKey

import haliax as hax
from haliax import Axis, NamedArray
from haliax.util import is_named_array
import pytest


def test_scan():
    Height = Axis("Height", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)
    named1 = hax.random.uniform(PRNGKey(0), (Height, Width, Depth))

    def scan_fun(acc, x):
        return acc + jnp.sum(x.array), x.take(Width, 2)

    total, selected = hax.scan(scan_fun, Height)(0.0, named1)

    assert jnp.all(jnp.isclose(total, jnp.sum(named1.array, axis=(0, 1, 2))))
    assert jnp.all(jnp.equal(selected.array, named1.take(Width, 2).array))

    total, selected = hax.scan(scan_fun, "Height")(0.0, named1)

    assert jnp.all(jnp.isclose(total, jnp.sum(named1.array, axis=(0, 1, 2))))
    assert jnp.all(jnp.equal(selected.array, named1.take("Width", 2).array))


def test_scan_not_0th_axis():
    Height = Axis("Height", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)
    named1 = hax.random.uniform(PRNGKey(0), (Height, Width, Depth))

    def scan_fun(acc, x):
        return acc + jnp.sum(x.array), x.take(Width, 2)

    total, selected = hax.scan(scan_fun, Depth)(0.0, named1)

    assert jnp.all(jnp.isclose(total, jnp.sum(named1.array, axis=(0, 1, 2))))
    assert jnp.all(jnp.equal(selected.array, named1.take(Width, 2).rearrange(selected.axes).array))


def test_scan_str_arg():
    Height = Axis("Height", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)
    named1 = hax.random.uniform(PRNGKey(0), (Height, Width, Depth))

    def scan_fun(acc, x):
        return acc + jnp.sum(x.array), x.take("Width", 2)

    total, selected = hax.scan(scan_fun, "Height")(0.0, named1)

    assert jnp.all(jnp.isclose(total, jnp.sum(named1.array, axis=(0, 1, 2))))
    assert jnp.all(jnp.equal(selected.array, named1.take(Width, 2).array))


def test_scan_static_args():
    Height = Axis("Height", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)
    named1 = hax.random.uniform(PRNGKey(0), (Height, Width, Depth))

    def scan_fun(acc, x, static1, *, static2):
        assert static1 is True
        assert static2 is False
        return acc + jnp.sum(x.array), x.take(Width, 2)

    total, selected = hax.scan(scan_fun, Depth, is_scanned=is_named_array)(0.0, named1, True, static2=False)

    assert jnp.all(jnp.isclose(total, jnp.sum(named1.array, axis=(0, 1, 2))))
    assert jnp.all(jnp.equal(selected.array, named1.take(Width, 2).rearrange(selected.axes).array))


def test_scan_doesnt_scan_scalars():
    Height = Axis("Height", 10)
    named1 = hax.random.uniform(PRNGKey(0), (Height,))

    def scan_fun(acc, z, x):
        return (acc + z * x).scalar(), x * z

    total, selected = hax.scan(scan_fun, Height)(0.0, 4.0, named1)

    assert jnp.all(jnp.isclose(total, jnp.sum(named1.array * 4.0)))
    assert jnp.all(jnp.equal(selected.array, named1.array * 4.0))


def test_scan_doesnt_scan_init():
    Height = Axis("Height", 10)
    named1 = hax.random.uniform(PRNGKey(0), (Height,))

    init = jnp.arange(Height.size, dtype=jnp.float32)

    def scan_fun(acc, z, x):
        out = acc + z * x, x * z
        return out

    total, selected = hax.scan(scan_fun, Height)(init, 4.0, named1)

    assert jnp.all(jnp.isclose(total, init + jnp.sum(named1.array * 4.0)))

    # double check with named array init
    total, selected = hax.scan(scan_fun, Height)(hax.named(init, "Height"), 4.0, named1)

    assert jnp.all(jnp.isclose(total.array, init + jnp.sum(named1.array * 4.0)))

    # now do fold
    def fold_fun(acc, z, x):
        return acc + z * x

    total = hax.fold(fold_fun, Height)(init, 4.0, named1)

    assert jnp.all(jnp.isclose(total, init + jnp.sum(named1.array * 4.0)))

    total = hax.fold(fold_fun, Height)(hax.named(init, "Height"), 4.0, named1)

    assert jnp.all(jnp.isclose(total.array, init + jnp.sum(named1.array * 4.0)))


def test_scan_hierarchical():
    Height = Axis("Height", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)
    named1 = hax.random.uniform(PRNGKey(0), (Height, Width, Depth))

    def scan_fun(acc, x):
        return acc + jnp.sum(x.array), x.take("Width", 2)

    total, selected = hax.scan(scan_fun, "Height")(0.0, named1)
    ckpt = hax.ScanCheckpointPolicy(nested=True)
    total_blocked, selected_blocked = hax.scan(scan_fun, "Height", remat=ckpt)(0.0, named1)

    assert jnp.all(jnp.isclose(total, total_blocked))
    assert jnp.all(jnp.equal(selected.array, selected_blocked.array))


def test_scan_reports_mismatched_unnamed_array():
    Height = Axis("Height", 2)

    def f(c, x, y):
        return c, x + y

    good = jnp.zeros((Height.size, 3))
    bad = jnp.zeros((Height.size - 1, 3))

    with pytest.raises(ValueError) as e:
        hax.scan(f, Height)(0, good, y=bad)

    assert "y has leading dimension" in str(e.value)


def test_scan_reports_eqx_module_field_path():
    Height = Axis("Height", 2)

    class Foo(eqx.Module):
        my_array: jnp.ndarray

    foo = Foo(jnp.zeros((Height.size - 1, 3)))

    def f(c, foo):
        return c, foo.my_array

    with pytest.raises(ValueError) as e:
        hax.scan(f, Height)(0, foo)

    assert "foo.my_array" in str(e.value)


def test_fold():
    Height = Axis("Height", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)
    named1 = hax.random.uniform(PRNGKey(0), (Height, Width, Depth))

    acc = hax.zeros((Height, Width))

    total = hax.fold(lambda x, y: x + y, Depth)(acc, named1)

    assert jnp.all(jnp.isclose(total.rearrange(acc.axes).array, jnp.sum(named1.array, axis=2)))


def test_fold_str_args():
    Height = Axis("Height", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)
    named1 = hax.random.uniform(PRNGKey(0), (Height, Width, Depth))

    acc = hax.zeros((Height, Width))

    total = hax.fold(lambda x, y: x + y, "Depth")(acc, named1)

    assert jnp.all(jnp.isclose(total.rearrange(acc.axes).array, jnp.sum(named1.array, axis=2)))


def test_fold_static_args():
    Height = Axis("Height", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)
    named1 = hax.random.uniform(PRNGKey(0), (Height, Width, Depth))

    def fold_fun(acc, x, static1, *, static2):
        assert static1 is True
        assert static2 is False
        return NamedArray(acc.array + x.rearrange(acc.axes).array, acc.axes)

    acc = hax.zeros((Height, Width))

    total = hax.fold(fold_fun, Depth)(acc, named1, True, static2=False)

    assert jnp.all(jnp.isclose(total.rearrange(acc.axes).array, jnp.sum(named1.array, axis=2)))


def test_fold_doesnt_reduce_scalars():
    Height = Axis("Height", 10)
    named1 = hax.random.uniform(PRNGKey(0), (Height,))

    acc = hax.zeros((Height,))

    total = hax.fold(lambda x, z, y: x + z * y, Height)(acc, 4.0, named1)

    assert jnp.all(jnp.isclose(total.rearrange(acc.axes).array, jnp.sum(named1.array * 4.0)))


def test_vmap_unmapped_args():
    Batch = Axis("Batch", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)
    named1 = hax.random.uniform(PRNGKey(0), (Width, Depth))

    def vmap_fun(x):
        return x.take(Width, 2)

    selected = hax.vmap(vmap_fun, Batch)(named1)

    expected_jax = jnp.array([named1.take(Width, 2).array for _ in range(Batch.size)])
    expected_names = (Batch, Depth)

    assert jnp.all(jnp.equal(selected.array, expected_jax))
    assert selected.axes == expected_names


def test_vmap_non_static_bool_fields():
    Batch = Axis("Batch", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)
    named1 = hax.random.uniform(PRNGKey(0), (Width, Depth))

    class Foo(eqx.Module):
        field: NamedArray
        flag: bool = False

        def __init__(self):
            super().__init__()
            self.field = named1
            self.flag = True

    vmap_foo = hax.vmap(Foo, Batch)()

    assert vmap_foo.field.axes == (Batch, Width, Depth)
    assert vmap_foo.flag is True


def test_vmap_mapped_args():
    Batch = Axis("Batch", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)
    named1 = hax.random.uniform(PRNGKey(0), (Batch, Width, Depth))

    def vmap_fun(x):
        return x.sum(Width)

    selected = hax.vmap(vmap_fun, Batch)(named1)

    expected_jax = jnp.array([named1.sum(Width).array for _ in range(Batch.size)])
    expected_names = (Batch, Depth)

    assert jnp.all(jnp.equal(selected.array, expected_jax))
    assert selected.axes == expected_names


def test_vmap_str_args():
    Batch = Axis("Batch", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)
    named1 = hax.random.uniform(PRNGKey(0), (Batch, Width, Depth))

    def vmap_fun(x):
        return x.sum(Width)

    selected = hax.vmap(vmap_fun, "Batch")(named1)

    expected_jax = jnp.array([named1.sum(Width).array for _ in range(Batch.size)])
    expected_names = (Batch, Depth)

    assert jnp.all(jnp.equal(selected.array, expected_jax))
    assert selected.axes == expected_names

    # also ensure that this works when we return a non-haliax array
    def vmap_fun2(x):
        return x.sum(Width).array

    selected = hax.vmap(vmap_fun2, "Batch")(named1)

    assert jnp.all(jnp.equal(selected, expected_jax))


def test_vmap_mapped_kwarg():
    Batch = Axis("Batch", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)
    named1 = hax.random.uniform(PRNGKey(0), (Batch, Width, Depth))

    def vmap_fun(x):
        return x.sum(Width)

    selected = hax.vmap(vmap_fun, Batch)(x=named1)

    expected_jax = jnp.array([named1.sum(Width).array for _ in range(Batch.size)])
    expected_names = (Batch, Depth)

    assert jnp.all(jnp.equal(selected.array, expected_jax))
    assert selected.axes == expected_names


def test_vmap_static_args():
    Batch = Axis("Batch", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)
    named1 = hax.random.uniform(PRNGKey(0), (Batch, Width, Depth))

    def vmap_fun(x, y):
        return x.sum(Width) if y else x

    selected = hax.vmap(vmap_fun, Batch)(named1, True)

    expected = hax.sum(named1, Width)

    assert jnp.all(jnp.equal(selected.array, expected.array))
    assert selected.axes == expected.axes


def test_vmap_error_for_incorrectly_specified_args():
    class Module(eqx.Module):
        # this should usually be declared static, but we're simulating a user error
        field: Axis

        def __call__(self, x):
            return x.sum(self.field)

    Batch = Axis("Batch", 10)
    Width = Axis("Width", 3)

    hax.vmap(lambda a: Module(a), Batch)(Width)


def test_vmap_multiple_axes():
    Batch1 = Axis("Batch1", 4)
    Batch2 = Axis("Batch2", 3)
    Width = Axis("Width", 2)
    Depth = Axis("Depth", 5)

    named = hax.random.uniform(PRNGKey(0), (Batch1, Batch2, Width, Depth))

    def vmap_fun(x):
        return x.sum(Width)

    selected = hax.vmap(vmap_fun, (Batch1, Batch2))(named)

    expected = jnp.sum(named.array, axis=2)

    assert jnp.allclose(selected.array, expected)
    assert selected.axes == (Batch1, Batch2, Depth)



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


# Support for einops-style rearrangement strings, but supporting named axes and unordered matching
import dataclasses
import typing
from types import EllipsisType
from typing import Mapping, Optional, Sequence

import jax.lax
import jax.numpy as jnp

from ..axis import Axis, AxisSelector, PartialAxisSpec, axis_name, rearrange_for_partial_order
from ..core import NamedArray
from ..partitioning import auto_sharded
from .parsing import AliasTable, Expression, _resolve_bindings, parse_rearrangement, raise_parse_error


@typing.overload
def rearrange(array: NamedArray, axes: Sequence[AxisSelector | EllipsisType]) -> NamedArray:
    pass


@typing.overload
def rearrange(array: NamedArray, expression: str, **bindings: AxisSelector | int) -> NamedArray:
    pass


def rearrange(array: NamedArray, *args, **kwargs) -> NamedArray:
    """
    Rearrange a tensor according to an einops-style haliax rearrangement string or a sequence of axes.
    See full documentation here: [Rearrange](https://haliax.readthedocs.io/en/latest/rearrange/)

    The sequence form of `rearrange` rearranges an array so that its underlying storage conforms to axes.
    axes may include up to 1 ellipsis, indicating that the remaining axes should be
    permuted in the same order as the array's axes.

    For example, if array has axes (a, b, c, d, e, f) and axes is (e, f, a, ..., c),
    then the output array will have axes (e, f, a, b, c, d).

    The string form of `rearrange` works similarly to einops.rearrange, but also supports named axes and unordered matching.
    The string form of `rearrange` comes in two forms:

    * **Ordered strings** are like einops strings, with the only significant difference that flattened axes are named with
    a colon, e.g. `B H W C -> B (E: H W C)`.
    * **Unordered strings** match axes by name and are marked with the addition of curly braces,
    e.g. `{H W C} -> ... C H W` or `{H W C} -> ... (E: H W C)`

    As with einops, you can provide axis sizes to unflatten axes. For instance, to turn an image patches,
    `hax.rearrange(x, '{ B (H: w1 H) (W: w1 W)} -> (B: B h1 w1) H W ...', H=32, W=32)
    will turn a batch of images into a batch of image patches. Bindings can also be [haliax.Axis][] objects,
    or strings that will be used as the actual name of the resulting axis.

    Examples:
        >>> import haliax as hax
        >>> import jax.random as jrandom
        >>> B, H, W, C = hax.Axis("B", 8), hax.Axis("H", 32), hax.Axis("W", 32), hax.Axis("C", 3)
        >>> x = hax.random.normal( (B, H, W, C))
        >>> # Sequence-based rearrange
        >>> hax.rearrange(x, (C, B, H, W))
        >>> hax.rearrange(x, (C, ...)) # ellipsis means "keep the rest of the axes in the same order"
        >>> # String-based rearrange
        >>> # permute the axes
        >>> hax.rearrange(x, "B H W C -> C B H W")
        >>> # flatten the image (note the assignment of a new name to the flattened axis)
        >>> hax.rearrange(x, "B H W C -> B (E: H W C)")
        >>> # turn the image into patches
        >>> hax.rearrange(x, "{ B (H: h1 H) (W: w1 W) C } -> (B: B h1 w1) (E: H W C) ...", H=2, W=2)
        >>> # names can be longer than one character
        >>> hax.rearrange(x, "{ B (H: h1 H) (W: w1 W) C } -> (B: B h1 w1) (embed: H W C) ...", H=2, W=2)
    """

    if len(args) == 1:
        axes = args[0]
        if isinstance(axes, str):
            return einops_rearrange(array, axes, **kwargs)
        else:
            return axis_spec_rearrange(array, axes)
    elif len(args) > 1:
        raise TypeError("Only one positional argument allowed")

    kwargs = dict(kwargs)
    expression = kwargs.pop("expression", None)
    if expression is not None:
        return einops_rearrange(array, expression, **kwargs)
    else:
        axes = kwargs.pop("axes", None)
        if axes is None:
            raise TypeError("Must specify either axes or expression")
        return axis_spec_rearrange(array, axes)


def axis_spec_rearrange(array: NamedArray, axis_spec: PartialAxisSpec) -> NamedArray:
    if len(axis_spec) == 0 and len(array.axes) != 0:
        raise ValueError("No axes specified")

    # various fast paths
    if len(axis_spec) == 1 and axis_spec[0] is Ellipsis:
        return array

    if axis_spec == array.axes:
        return array

    if axis_spec[-1] is Ellipsis and array.axes[0 : len(axis_spec) - 1] == axis_spec[0 : len(axis_spec) - 1]:
        return array

    if axis_spec[0] is Ellipsis and array.axes[len(axis_spec) - 1 :] == axis_spec[1:]:
        return array

    out_axes = rearrange_for_partial_order(axis_spec, array.axes)

    # now build a permute_spec
    permute_spec = []
    index_of_in = {ax: i for i, ax in enumerate(array.axes)}

    for ax in out_axes:
        permute_spec.append(index_of_in[ax])

    out_axes = tuple(array.axes[i] for i in typing.cast(list[int], permute_spec))
    return NamedArray(jnp.transpose(array.array, permute_spec), out_axes)


def einops_rearrange(array: NamedArray, expression: str, **bindings: AxisSelector | int) -> NamedArray:
    lhs, rhs = parse_rearrangement(expression)

    # the fundamental xla op for rearranging is reshape, which combines both transpose and reshape
    # all rearranges are fundamentally a reshape, followed by a transpose, followed by a reshape
    # as an example of an op that needs both, consider:
    #  '(a b) c d -> (c a) (b d)'
    # where a = 2, b = 3, c = 4, d = 5
    # which is equivalent to:
    # x.reshape((2, 3, 4, 5)).transpose((2, 0, 1, 3)).reshape((4 * 2, 3 * 5))
    plan = _plan_rearrange(expression, lhs, rhs, array, bindings)

    raw_array = array.array

    if plan.intermediate_axes != array.axes:
        raw_array = raw_array.reshape([ax.size for ax in plan.intermediate_axes])
        array = NamedArray(raw_array, plan.intermediate_axes)
        array = auto_sharded(array)
        raw_array = array.array

    final_shape = tuple(ax.size for ax in plan.final_axes)

    if plan.needs_final_reshape:
        finished_array = jax.lax.reshape(raw_array, new_sizes=final_shape, dimensions=plan.transpose)
    elif plan.transpose is not None:
        finished_array = jax.lax.transpose(raw_array, permutation=plan.transpose)
    else:
        finished_array = raw_array

    return auto_sharded(NamedArray(finished_array, plan.final_axes))


@dataclasses.dataclass(frozen=True)
class _Plan:
    intermediate_axes: tuple[Axis, ...]
    transpose: Optional[tuple[int, ...]]
    needs_final_reshape: bool

    final_axes: tuple[Axis, ...]


def _plan_rearrange(
    original_str, lhs: Expression, rhs: Expression, array: NamedArray, input_bindings: Mapping[str, AxisSelector | int]
) -> _Plan:
    aliases = _resolve_bindings(array, input_bindings)
    grouped_new_shapes = _determine_initial_reshape(original_str, lhs, array, aliases)
    intermediate_axes = tuple(ax for split_axes in grouped_new_shapes for ax in split_axes)

    transpose: Optional[tuple[int, ...]]
    transpose, final_axes = _determine_final_transpose_and_reshape(original_str, rhs, aliases, intermediate_axes)

    transposed_intermediate_axes = tuple(intermediate_axes[i] for i in transpose)
    if transposed_intermediate_axes == final_axes:
        needs_final_reshape = False
    else:
        needs_final_reshape = True

    if transpose == tuple(range(len(transpose))):
        transpose = None

    return _Plan(intermediate_axes, transpose, needs_final_reshape, final_axes)


def _determine_final_transpose_and_reshape(
    expression, rhs: Expression, aliases: "AliasTable", intermediate_axes: tuple[Axis, ...]
) -> tuple[tuple[int, ...], tuple[Axis, ...]]:
    # The rhs tells us three things:
    # 1. how to reorder the intermediate axes
    # 2. how to merge the intermediate axes into the final axes
    # 3. where ellipses are in the final axes (where we can put unused intermediate axes)

    # MUTATES `aliases`

    # Our approach is to:
    # 1. Figure out the partial order of the intermediate axes
    # 2. Compute the full final axes (using rearrange_for_partial_order)
    # 3. Figure out the transposition order to get the intermediate axes into the right order
    # return (2) and (3)

    transposed_intermediate_axes_order: list[Axis | EllipsisType] = []

    first_intermediate_for_final: dict[Axis, Axis] = {}  # intermediate -> final

    used_intermediate_axes = set()  # axes must be used exactly once
    has_ellipsis = False

    for cpos, capture in enumerate(rhs.captures):
        if capture == Ellipsis:
            has_ellipsis = True
            transposed_intermediate_axes_order.append(Ellipsis)
            continue

        assert not isinstance(capture, EllipsisType)

        binding = capture.binding
        if binding is None:
            raise_parse_error("All rhs axes must have a name. Use (name: bindings)", expression, capture.char_range)

        sz = 1
        first_axis = None
        for ax_name in capture.axes:
            intermed_axis = aliases.dealias_binding(ax_name)

            if not isinstance(intermed_axis, Axis):
                raise_parse_error(f"Axis {ax_name} is not bound on the lhs", expression, capture.char_range)

            if first_axis is None:
                first_axis = intermed_axis

            if intermed_axis in used_intermediate_axes:
                raise_parse_error(f"Axis {intermed_axis} is used more than once", expression, capture.char_range)
            used_intermediate_axes.add(intermed_axis)
            transposed_intermediate_axes_order.append(intermed_axis)

            sz *= intermed_axis.size
            # now find the position of this axis in the intermediate axes

        # figure out the name of the final axis
        axis = aliases.dealias_binding(binding)
        if axis is None:
            new_axis = Axis(binding, sz)
        else:
            new_axis = Axis(axis_name(axis), sz)

        # TODO: this isn't ideal for the unusual case where we have an empty set of axes to bind to an axis
        # we won't accurately get the order for these unitary axes
        if first_axis is not None:
            first_intermediate_for_final[first_axis] = new_axis

    if not has_ellipsis:
        unused_intermediate_axes = set(intermediate_axes) - used_intermediate_axes
        if len(unused_intermediate_axes) > 0:
            raise ValueError("Not all intermediate axes are used. Use ... to insert unused axes")

    # now we have a partial order of the intermediate axes
    reordered_intermediate_axes = rearrange_for_partial_order(transposed_intermediate_axes_order, intermediate_axes)

    # now we need to figure out the final axes
    final_axes: list[Axis] = []
    transposition_order: list[int] = []

    for intermediate_axis in reordered_intermediate_axes:
        if intermediate_axis not in used_intermediate_axes:
            # this axis is not used, so we keep it from the reordered axes
            final_axes.append(intermediate_axis)
            transposition_order.append(intermediate_axes.index(intermediate_axis))
        else:
            final_axis = first_intermediate_for_final.get(intermediate_axis, None)

            if final_axis is not None:
                final_axes.append(final_axis)

            transposition_order.append(intermediate_axes.index(intermediate_axis))

    return tuple(transposition_order), tuple(final_axes)  # type: ignore


def _determine_initial_reshape(
    expression,
    lhs: Expression,
    array: NamedArray,
    aliases: AliasTable,
) -> list[list[Axis]]:
    # MUTATES `aliases`
    # the lhs all need to be bound to axes in the array, or synthesized as parts of axes.
    # In the lhs, bindings look like either a name, or a name and a list of (new) axes.
    # bindings can either be done by name, or by position, depending on if lhs.is_ordered
    new_shapes: list[Optional[list[Axis]]] = [None] * len(array.axes)
    used_new_names: set[str] = set()  # names can only be used once on a side

    # one subtle difference between the lhs and the rhs is the handling of binding in expressions like (a: b c)
    # in the lhs, this means that a is bound to an axis, and b and c are split out from a
    # in the rhs, this means that b and c must already be bound to axes, and a is a new axis name
    # this makes sense in the same way that pattern matching works

    if lhs.is_ordered:
        # if we start with an ellipsis, we bind from the right
        # if we end with an ellipsis, we bind from the left
        ellipsis_pos = None
        axis_index_for_capture: list[Optional[int]] = [None] * len(lhs.captures)
        covered_axes = set()
        # bind from the left
        axis_pos = 0
        for cpos, capture in enumerate(lhs.captures):
            if capture == Ellipsis:
                if ellipsis_pos is not None:
                    assert False, "should not be here"  # pragma: no cover
                ellipsis_pos = cpos
                break

            assert not isinstance(capture, EllipsisType)

            if axis_pos >= len(array.axes):
                raise_parse_error("Too many axes in lhs", expression, capture.char_range)

            axis_index_for_capture[cpos] = axis_pos
            covered_axes.add(axis_pos)
            axis_pos += 1

        # bind from the right, take care to check second ellipsis
        if ellipsis_pos is not None:
            axis_pos = len(array.axes) - 1
            for cpos in range(len(lhs.captures) - 1, ellipsis_pos, -1):
                capture = lhs.captures[cpos]
                if capture == Ellipsis:
                    raise_parse_error("Only one ellipsis allowed", expression, None)

                assert not isinstance(capture, EllipsisType)

                axis_index_for_capture[cpos] = axis_pos
                if axis_pos in covered_axes:
                    raise_parse_error(
                        f"Axis {array.axes[axis_pos]} is bound more than once",
                        expression,
                        capture.char_range,
                    )
                covered_axes.add(axis_pos)
                axis_pos -= 1
        else:
            # no ellipsis, so we need to check that we covered all axes
            if len(covered_axes) < len(array.axes):
                raise_parse_error(
                    "Not all axes are bound, use ... to skip missing axes", expression, len(expression) - 1
                )  # type: ignore
            elif len(covered_axes) > len(array.axes):
                raise_parse_error("Too many axes are bound", expression, lhs.captures[-1].char_range)  # type: ignore

        # now that we have the bindings, we can figure out the new shapes
        for cpos, capture in enumerate(lhs.captures):
            if capture == Ellipsis:
                continue
            assert not isinstance(capture, EllipsisType)

            axis_index = axis_index_for_capture[cpos]
            if axis_index is None:
                raise_parse_error("Internal error", expression, capture.char_range)

            axis = array.axes[axis_index]
            if new_shapes[axis_index] is not None:
                raise_parse_error(f"Axis {axis} is bound more than once", expression, capture.char_range)

            new_axes = _solve_split_axes(axis, capture, aliases, used_new_names, expression)
            new_shapes[axis_index] = new_axes  # type: ignore

    else:
        # we just need to bind the axes in the lhs to the axes in the array
        for capture in lhs.captures:
            # ellipses are ignored in unordered rearrangements
            if capture == Ellipsis:
                continue

            assert not isinstance(capture, EllipsisType)

            if capture.binding is None:
                raise_parse_error(
                    "Unordered axes must be bound by name, e.g. (a: b) or just a", expression, capture.char_range
                )

            # let's see if we're aliasing it in the bindings
            maybe_alias = aliases.dealias_binding(capture.binding)
            if maybe_alias is None:
                maybe_alias = capture.binding
            else:
                maybe_alias = axis_name(maybe_alias)

            try:
                axis = array.resolve_axis(maybe_alias)
                # aliases.bind_alias(capture.binding, axis, expression, capture.char_range)
            except ValueError:
                raise_parse_error(f"Could not resolve axis {maybe_alias}", expression, capture.char_range)

            try:
                axis_index = array.axes.index(axis)
            except ValueError:
                raise_parse_error(f"Axis {axis} is not in the array", expression, capture.char_range)
            if new_shapes[axis_index] is not None:
                raise_parse_error(f"Axis {axis} is bound more than once", expression, capture.char_range)

            new_axes = _solve_split_axes(axis, capture, aliases, used_new_names, expression)
            new_shapes[axis_index] = new_axes  # type: ignore

    for i in range(len(array.axes)):
        if new_shapes[i] is None:
            new_shapes[i] = [array.axes[i]]

    return new_shapes  # type: ignore


def _solve_split_axes(axis, capture, aliases, used_new_names, expression):
    """
    Given an axis and a capture of the form (a: b c) or (b c) on the lhs, solve for the new axes.
    """
    new_axes: list[Optional[Axis]] = []
    unsolved_axis_index: Optional[int] = None

    # easy case: 1 axis in capture
    if len(capture.axes) == 1:
        new_axis_name = capture.axes[0]
        if new_axis_name in used_new_names:
            raise_parse_error(f"Capture {new_axis_name} is assigned more than once", expression, capture.char_range)
        used_new_names.add(new_axis_name)

        new_axes.append(axis)
        aliases.bind_alias(new_axis_name, axis, expression, capture.char_range)

        return new_axes

    remaining_size = axis.size

    for new_axis_name in capture.axes:
        if new_axis_name in used_new_names:
            raise_parse_error(
                f"Capture {new_axis_name} is assigned more than once in lhs", expression, capture.char_range
            )

        used_new_names.add(new_axis_name)

        new_axis = aliases.dealias_binding(new_axis_name)
        if new_axis is None:
            new_axis = new_axis_name

        if isinstance(new_axis, Axis):
            new_axes.append(new_axis)
            if remaining_size % new_axis.size:
                raise_parse_error(f"Axes do not divide evenly into axis {axis}", expression, capture.char_range)
            remaining_size //= new_axis.size
        else:
            if unsolved_axis_index is not None:
                raise_parse_error(
                    "Sizes for this split axis are ambiguous. You must provide a size as a kwarg.",
                    expression,
                    capture.char_range,
                )
            unsolved_axis_index = len(new_axes)
            new_axes.append(None)

    if unsolved_axis_index is not None:
        # we need to solve for this axis
        unsolved_alias = aliases.dealias_binding(capture.axes[unsolved_axis_index])
        assert not isinstance(unsolved_alias, Axis)
        if unsolved_alias is None:
            unsolved_alias = capture.axes[unsolved_axis_index]
        assert isinstance(unsolved_alias, str)

        new_axis = Axis(unsolved_alias, remaining_size)
        new_axes[unsolved_axis_index] = new_axis
        aliases.bind_alias(capture.axes[unsolved_axis_index], new_axis, expression, capture.char_range)
    return new_axes



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


import dataclasses
from types import EllipsisType
from typing import Mapping, NoReturn, Optional, Sequence

from haliax.axis import Axis, AxisSelector


@dataclasses.dataclass(frozen=True)
class _AxisCapture:
    binding: Optional[str] = None
    axes: tuple[str, ...] = ()
    char_range: Optional[tuple[int, int]] = None

    def __post_init__(self):
        if len(self.axes) == 0:
            raise ValueError("Empty axes not allowed")


@dataclasses.dataclass(frozen=True)
class Expression:
    captures: Sequence[_AxisCapture | EllipsisType]
    is_ordered: bool


def raise_parse_error(message: str, expression: str, pos: Optional[int | tuple[int, int]]) -> NoReturn:
    """Raise a ValueError with a message and the position in the expression."""
    fmt = f"Error while parsing:\n    {expression}"
    if pos is not None:
        if isinstance(pos, int):
            fmt += f'\n    {" " * pos}^'
        else:
            fmt += f"\n    {' ' * pos[0]}{'^' * max(1, pos[1] - pos[0])}"

    fmt += f"\n{message}"

    raise ValueError(fmt)


def _parse_quoted_string(expression: str, pos: int) -> tuple[str, int]:
    """Parse a quoted string from an einops-style haliax rearrangement string."""

    if expression[pos] not in "'\"":
        raise_parse_error(f"Expected \" or ' at position {pos}", expression, pos)
    quote = expression[pos]
    pos += 1
    ident = ""
    while pos < len(expression):
        if expression[pos] == quote:
            pos += 1
            break
        elif expression[pos] == "\\":
            pos += 1
            if pos >= len(expression):
                raise_parse_error(f"Unexpected end of string at position {pos}", expression, pos)
            ident += expression[pos]
            pos += 1
            continue
        else:
            ident += expression[pos]
            pos += 1
            continue
    if len(ident) == 0:
        raise_parse_error("Empty strings are not valid identifiers", expression, pos)

    return ident, pos


def _parse_ident(expression: str, pos: int) -> tuple[str, int]:
    """parses an identifier or string literal from an einops-style haliax rearrangement string."""
    if expression[pos] in "'\"":
        return _parse_quoted_string(expression, pos)
    else:
        ident = ""
        while pos < len(expression):
            if str.isalnum(expression[pos]) or expression[pos] == "_":
                if len(ident) == 0 and str.isdigit(expression[pos]):
                    raise_parse_error("Identifiers cannot start with a number", expression, pos)
                ident += expression[pos]
                pos += 1
                continue
            else:
                break
        if len(ident) == 0:
            raise_parse_error("Identifier expected", expression, pos)

        return ident, pos


def _parse_group(expression, pos):
    # parses a group of axes like (a b c) or (a: b c)
    pos_in = pos
    if expression[pos] != "(":
        raise ValueError("Expected (")
    pos += 1
    binding = None
    axes: list[str] = []
    current_ident = ""
    while pos < len(expression):
        if expression[pos] == ")":
            pos += 1
            break
        elif expression[pos] == ":":
            if binding is not None:
                raise_parse_error("Only one binding allowed per group", expression, pos)
            if not current_ident:
                raise_parse_error("Binding cannot be empty", expression, pos)
            if len(axes) > 0:
                raise_parse_error("Binding must come before axes", expression, pos)
            binding = current_ident
            current_ident = ""
            pos += 1
            continue
        elif str.isspace(expression[pos]):
            if current_ident:
                axes.append(current_ident)
                current_ident = ""
            pos += 1
            continue
        elif expression[pos] == "(":
            raise_parse_error("Only one level of nesting is allowed", expression, pos)
        elif expression[pos] == "}":
            raise ValueError(f"Unexpected }} at {pos}")
        elif str.isalnum(expression[pos]) or expression[pos] == "_":
            # don't allow numbers at the start of an identifier
            if len(current_ident) == 0 and str.isdigit(expression[pos]):
                raise_parse_error("Identifiers cannot start with a number", expression, pos)
            current_ident += expression[pos]
            pos += 1
            continue
        elif expression[pos] in "'\"":
            # parse quoted string as identifier
            if current_ident:
                axes.append(current_ident)

            ident, pos = _parse_quoted_string(expression, pos)
            current_ident = ident
            continue
        else:
            raise_parse_error(f"Unexpected character {expression[pos]}", expression, pos)

    if current_ident:
        axes.append(current_ident)

    if len(axes) == 0:
        raise_parse_error("No axes found", expression, pos_in)

    # todo: should we allow anonymous/literal
    char_range = (pos_in, pos)
    return _AxisCapture(binding, tuple(axes), char_range), pos


def _parse_expression(expression: str, pos) -> tuple[Expression, int]:
    """Parse one side of an einops-style haliax rearrangement string."""
    captures = []
    is_ordered = True
    seen_char = False
    finished = False

    while pos < len(expression):
        if expression[pos] == "{":
            if seen_char:
                raise_parse_error("Unexpected {", expression, pos)
            seen_char = True
            is_ordered = False
            pos += 1
            continue
        elif expression[pos] == "}":
            if is_ordered:
                raise_parse_error("Unexpected }", expression, pos)
            pos += 1
            finished = True
            continue
        elif expression[pos] == "(":
            if finished:
                raise_parse_error("Unexpected ( after }", expression, pos)
            seen_char = True
            capture, pos = _parse_group(expression, pos)
            captures.append(capture)
            continue
        elif str.isspace(expression[pos]):
            pos += 1
            continue
        elif expression[pos : pos + 3] == "...":
            seen_char = True
            if finished:
                raise_parse_error("Unexpected ... after }", expression, pos)
            captures.append(Ellipsis)
            pos += 3
            continue
        elif expression[pos] == "-":
            # if not seen_char:
            #     _raise_error("Unexpected -", expression, pos)
            if pos + 1 >= len(expression):
                raise_parse_error("Unexpected end of string", expression, pos)
            if expression[pos + 1] != ">":
                raise_parse_error("Expected >", expression, pos)
            break
        elif expression[pos] == ",":
            # this ends the current expression if we're not inside a group or unordered capture
            if not is_ordered:
                # treat as whitespace
                pos += 1
                continue
            if finished:
                break

            break
        else:
            if finished:
                raise_parse_error("Unexpected character after }", expression, pos)
            ident, new_pos = _parse_ident(expression, pos)
            captures.append(_AxisCapture(binding=ident, axes=(ident,), char_range=(pos, new_pos)))
            seen_char = True
            pos = new_pos
            continue

    if not finished and not is_ordered:
        raise_parse_error("Expected }", expression, pos)

    return Expression(captures, is_ordered), pos


class AliasTable:
    bindings: dict[str, AxisSelector]  # names in the string to axes

    def __init__(self, bindings=None):
        if bindings is None:
            self.bindings = {}
        else:
            self.bindings = {**bindings}

    def dealias_binding(self, binding: str) -> Optional[AxisSelector]:
        return self.bindings.get(binding, None)

    def bind_alias(self, alias: str, axis: Axis, expr, char_range):
        if axis.name in self.bindings:
            if self.bindings[alias] != axis:
                raise_parse_error(f"Alias {alias} is assigned to more than one axis", expr, char_range)
        elif alias in self.bindings:
            current = self.bindings[alias]
            if isinstance(current, Axis):
                if current != axis:
                    raise_parse_error(f"Alias {alias} is assigned to more than one axis", expr, char_range)
            elif current != axis.name:
                raise_parse_error(f"Alias {alias} is assigned to more than one axis", expr, char_range)
            else:
                self.bindings[alias] = axis
        else:
            self.bindings[alias] = axis


def _resolve_bindings(array, bindings: Mapping[str, Axis | str | int]) -> AliasTable:
    b: dict[str, AxisSelector] = {}
    for name, selector in bindings.items():
        if isinstance(selector, str):
            try:
                selector = array.resolve_axis(selector)
            except ValueError:
                pass
        elif isinstance(selector, int):
            selector = Axis(name, selector)
        assert not isinstance(selector, int)
        b[name] = selector
    return AliasTable(b)


def parse_rearrangement(expression: str) -> tuple[Expression, Expression]:
    """Parse an einops-style haliax rearrangement string."""
    pos = 0
    lhs, pos = _parse_expression(expression, pos)

    # consume the ->
    if pos + 2 >= len(expression):
        raise_parse_error("Unexpected end of string", expression, pos)
    if expression[pos : pos + 2] != "->":
        raise_parse_error("Expected ->", expression, pos)

    pos += 2
    rhs, pos = _parse_expression(expression, pos)

    # make sure we consumed the whole string
    if pos != len(expression):
        raise_parse_error("Unexpected character", expression, pos)

    return lhs, rhs


def parse_einsum(expression: str) -> tuple[Sequence[Expression], Expression]:
    """
    Parse a haliax-style einsum string. This is an expansion of the einsum syntax
    to support named axes
    """
    pos = 0
    lhses = []
    while pos < len(expression):
        lhs, pos = _parse_expression(expression, pos)
        lhses.append(lhs)
        if pos >= len(expression):
            break
        if expression[pos] == ",":
            pos += 1
            continue
        elif expression[pos] == "-":
            if pos + 1 >= len(expression):
                raise_parse_error("Unexpected end of string", expression, pos)
            if expression[pos + 1] != ">":
                raise_parse_error("Expected >", expression, pos)
            pos += 2
            break
        else:
            raise_parse_error("Expected , or ->", expression, pos)

    rhs, pos = _parse_expression(expression, pos)

    if any(x is None for x in lhses):
        if len(lhses) > 1:
            raise_parse_error("If there are multiple lhs expressions, they must all be ordered", expression, pos)
        lhses = None  # type: ignore

    if pos != len(expression):
        raise_parse_error("Unexpected character", expression, pos)

    return lhses, rhs



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


import jax
import jax.random as jrandom
from jax import numpy as jnp

import haliax as hax
from haliax.nn import MoELinear


def _expected_moe_linear_output(moe: MoELinear, x: hax.NamedArray, group_sizes: hax.NamedArray):
    dim_numbers = jax.lax.RaggedDotDimensionNumbers(
        (
            ((x.axis_indices(moe.In),), (moe.weight.axis_indices(moe.In),)),
            ((), ()),
        ),
        x.axis_indices(hax.axis.without_axes(x.axes, moe.In)),
        (moe.weight.axis_indices(moe.Experts),),
    )
    out_raw = jax.lax.ragged_dot_general(
        lhs=x.array,
        rhs=moe.weight.array,
        group_sizes=group_sizes.array,
        ragged_dot_dimension_numbers=dim_numbers,
    )
    out_axes = hax.replace_axis(x.axes, moe.In, moe.Out)
    out = hax.named(out_raw, out_axes)
    if moe.bias is not None:
        out = out + moe.bias
    return out


def test_moe_linear_matches_ragged_dot_general():
    B, In, Out, E = hax.make_axes(B=3, In=4, Out=5, E=2)
    key = jrandom.PRNGKey(0)
    moe = MoELinear.init(E, In, Out, key=key)

    x = hax.random.normal(jrandom.PRNGKey(1), (B, In))
    group_sizes = hax.named(jnp.array([2, 1], dtype=jnp.int32), (E,))

    actual = moe(x, group_sizes)
    expected = _expected_moe_linear_output(moe, x, group_sizes)

    assert actual.axes == expected.axes
    assert jnp.allclose(actual.array, expected.array, rtol=1e-5, atol=1e-5)


def test_moe_linear_out_first_property():
    E, In, Out = hax.make_axes(E=2, In=4, Out=3)
    moe = MoELinear.init(E, In, Out, key=jrandom.PRNGKey(0), out_first=True)
    assert moe.out_first
    assert moe.weight.axes[:3] == (E, Out, In)

    moe2 = MoELinear.init(E, In, Out, key=jrandom.PRNGKey(1), out_first=False)
    assert not moe2.out_first
    assert moe2.weight.axes[:3] == (E, In, Out)


def test_moe_linear_gmm_matches_ragged_dot_general():
    B, In, Out, E = hax.make_axes(B=3, In=4, Out=5, E=2)
    moe = MoELinear.init(E, In, Out, key=jrandom.PRNGKey(0), use_gmm=True)

    x = hax.random.normal(jrandom.PRNGKey(1), (B, In))
    group_sizes = hax.named(jnp.array([2, 1], dtype=jnp.int32), (E,))

    with jax.sharding.Mesh(jax.devices(), ("data",)):
        actual = moe(x, group_sizes)

    expected = _expected_moe_linear_output(moe, x, group_sizes)

    assert actual.axes == expected.axes
    assert jnp.allclose(actual.array, expected.array, rtol=1e-5, atol=1e-5)



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


# Module to support torch-style "state dict" serialization via safetensors
import dataclasses
import typing
from typing import Any, Optional, Sequence, TypeVar

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
from jax import ShapeDtypeStruct
from jax.experimental.multihost_utils import sync_global_devices
from jax.sharding import Mesh, NamedSharding, PartitionSpec
from jax.tree_util import DictKey, FlattenedIndexKey, GetAttrKey, SequenceKey
from jaxtyping import PyTree

import haliax.partitioning as partitioning
from haliax._src.util import index_where
from haliax.core import NamedArray, named
from haliax.jax_utils import is_jax_array_like, is_scalarish
from haliax.tree_util import scan_aware_tree_map

try:
    import safetensors
except ImportError:
    safetensors = None


StateDict = dict[str, Any]
Mod = TypeVar("Mod", bound=eqx.Module)
T = TypeVar("T")


def from_torch_compatible_state_dict(
    t: T, state_dict: StateDict, *, unflatten: bool = True, prefix: Optional[str] = None
) -> T:
    """
    Convert a state dict to a tree that is compatible with the structure of `t`.

    If unflatten is true, then the weights in the state dict are assumed to have been flattened (as by flatten_modules_for_export).

    This applies [haliax.state_dict.from_state_dict][]
    """
    if unflatten:
        t = _flatten_to_unflatten(t, state_dict, prefix)
    else:
        t = from_state_dict(t, state_dict, prefix=prefix)

    return t


def flatten_modules_for_export(t: T) -> T:
    """
    Flatten all modules in a tree for export to torch.
    """

    def _flatten_module(module):
        if isinstance(module, ModuleWithStateDictSerialization):
            module = module.flatten_for_export()
            module = scan_aware_tree_map(
                _flatten_module,
                module,
                is_leaf=lambda x: x is not module and isinstance(x, ModuleWithStateDictSerialization),
            )
        return module

    return scan_aware_tree_map(_flatten_module, t, is_leaf=lambda x: isinstance(x, ModuleWithStateDictSerialization))


def unflatten_modules_from_export(t: T, template: T) -> T:
    """
    Unflatten all modules in a tree after import from torch.
    """

    def _unflatten_module(module, template):
        if isinstance(module, ModuleWithStateDictSerialization):
            module = module.unflatten_from_export(template)
            module = scan_aware_tree_map(
                _unflatten_module,
                module,
                template,
                is_leaf=lambda x: x is not module and isinstance(x, ModuleWithStateDictSerialization),
            )
        return module

    return scan_aware_tree_map(
        _unflatten_module, t, template, is_leaf=lambda x: isinstance(x, ModuleWithStateDictSerialization)
    )


def _flatten_to_unflatten(t, state_dict, prefix):
    """
    Flatten the torch compatible state_dict before loading into t, and then recover the unflattened layers.
    """

    # typically, `t` is a bunch of ShapeDtypeStructs, which can't be transposed etc. so we instead have to zeros()
    # into real arrays (that aren't actually real b/c this is inside a jit)
    def _dt_struct_to_array(struct):
        if not isinstance(struct, ShapeDtypeStruct):
            return struct
        return jnp.zeros(struct.shape, struct.dtype)

    t = jax.tree.map(_dt_struct_to_array, t)
    flat_t = flatten_modules_for_export(t)
    flat_t = from_state_dict(flat_t, state_dict, prefix=prefix)
    t = unflatten_modules_from_export(flat_t, t)
    return t


@typing.overload
def with_prefix(prefix: str | None, leaf: str) -> str: ...


@typing.overload
def with_prefix(prefix: str, leaf: None) -> str: ...


@typing.overload
def with_prefix(prefix: Optional[str], leaf: Optional[str]) -> Optional[str]: ...


def with_prefix(prefix: Optional[str], leaf: Optional[str]) -> Optional[str]:
    """Joins two optional path strings in a way compatible with pytorch state dict serialization"""
    if prefix is None:
        return leaf
    elif leaf is None:
        return prefix
    else:
        return f"{prefix}.{leaf}"


class ModuleWithStateDictSerialization(eqx.Module):
    """An eqx.Module that can be serialized to a torch-style state dict."""

    def to_state_dict(self, prefix: Optional[str] = None) -> StateDict:
        return default_eqx_module_to_state_dict(self, prefix)

    def from_state_dict(self: Mod, state_dict: StateDict, prefix: Optional[str] = None) -> Mod:
        return default_eqx_module_from_state_dict(self, state_dict, prefix)

    def _state_dict_key_map(self) -> dict[str, Optional[str]]:
        """Returns a dict mapping eqx.Module keys to torch keys that need to be renamed for serialization"""
        return {}

    def flatten_for_export(self: Mod) -> Mod:
        """
        Flatten articulated named arrays for export to torch. In general this method should, for a linear layer, flatten
        all input axes into a single axis, and all output axes into a single axis. You can do whatever else
        you want to support pytorch-compatible serialization if you want.

        This method is less general than to_state_dict and is only called when using to_torch_compatible_state_dict.
        """
        return self

    def unflatten_from_export(self: Mod, template: Mod) -> Mod:
        """
        Unflatten the module after import from torch.

        Template has the proper structure (e.g. articulated named axes) but the values are meaningless.
        """
        del template
        return self


def from_state_dict(tree: T, state_dict: StateDict, prefix: Optional[str] = None) -> T:
    """
    Given a (template) tree and a state dict, return a new tree with the same structure as the input tree, but with
    the values from the state dict.

    Args:
        tree: The template tree
        state_dict: The state dict
        prefix: The prefix to use when looking up keys in the state dict

    Returns:
        A new tree with the same structure as the input tree, but with the values from the state dict.

    """
    # TODO: assert compatibility of old and new values (type, shape, etc.)
    if isinstance(tree, eqx.Module):
        if hasattr(tree, "from_state_dict"):
            return tree.from_state_dict(state_dict, prefix)
        else:
            return default_eqx_module_from_state_dict(tree, state_dict, prefix)
    elif isinstance(tree, list):
        return [from_state_dict(item, state_dict, with_prefix(prefix, str(i))) for i, item in enumerate(tree)]  # type: ignore
    elif isinstance(tree, dict):
        return {k: from_state_dict(v, state_dict, prefix=with_prefix(prefix, k)) for k, v in tree.items()}  # type: ignore
    elif isinstance(tree, NamedArray):
        if prefix is None:
            raise ValueError("Cannot extract a leaf value from a torch dict without a prefix")

        array = state_dict[prefix]

        if isinstance(array, np.ndarray):
            mesh = partitioning._get_mesh()
            # TODO: modernize this
            if jax.device_count() > 1:  # this happens with the default mesh
                pspec = partitioning.pspec_for_axis(tree.axes)
                sharding = jax.sharding.NamedSharding(mesh, pspec)
                array = jax.make_array_from_callback(tree.array.shape, sharding, lambda indices: array[indices])
            else:
                array = jnp.array(array)
            array = named(array, tree.axes)
        else:
            array = named(array, tree.axes)
            array = partitioning.auto_sharded(array)

        return array  # type: ignore
    elif is_jax_array_like(tree):
        if prefix is None:
            raise ValueError("Cannot extract a leaf value from a state dict without a prefix")
        # TODO: add "strict" flag so we can return None in cases where it's just missing
        return jnp.array(state_dict[prefix])
    elif tree is None:
        return None  # type: ignore
    else:
        if prefix is None:
            return tree
        return state_dict.get(prefix, tree)


def to_state_dict(tree: PyTree, prefix: Optional[str] = None) -> StateDict:
    """
    Convert a PyTree to a state dict.

    Returns:
        The state dict representation of the input tree.
    """
    if tree is None:
        return {}
    elif isinstance(tree, eqx.Module):
        if hasattr(tree, "to_state_dict"):
            state_dict = tree.to_state_dict(prefix)
        else:
            state_dict = default_eqx_module_to_state_dict(tree, prefix)
    elif isinstance(tree, list):
        state_dict = {}
        for i, item in enumerate(tree):
            child = to_state_dict(item, prefix=with_prefix(prefix, str(i)))
            # TODO: check for conflicts?
            state_dict.update(child)
    elif isinstance(tree, dict):
        state_dict = {}
        for k, v in tree.items():
            child = to_state_dict(v, prefix=with_prefix(prefix, k))
            # TODO: check for conflicts?
            state_dict.update(child)
    elif isinstance(tree, NamedArray):
        if prefix is None:
            raise ValueError("Cannot convert a leaf value to a state dict without a prefix")
        if tree.array is not None:
            state_dict = {prefix: tree.array}
        else:
            state_dict = {}
    elif is_jax_array_like(tree):
        if prefix is not None:
            if tree is not None:
                state_dict = {prefix: tree}
            else:
                state_dict = {}
        else:
            raise ValueError("Cannot convert a leaf value to a state dict without a prefix")
    elif is_scalarish(tree):
        if prefix is not None:
            state_dict = {prefix: tree}
        else:
            raise ValueError("Cannot convert a leaf value to a state dict without a prefix")
    else:
        raise ValueError(f"Unsupported type {type(tree)}")

    return state_dict


def default_eqx_module_from_state_dict(mod: Mod, state_dict: StateDict, prefix: Optional[str] = None) -> Mod:
    key_map: Dict[str, Optional[str]] = getattr(mod, "_state_dict_key_map", lambda: {})()  # type: ignore
    names = []
    values = []
    for field in dataclasses.fields(mod):
        # only encode dynamic fields (i.e. fields that don't have the static flag)
        if field.metadata.get("static", False):
            continue
        key = key_map.get(field.name, field.name)
        value = getattr(mod, field.name)
        # TODO: might want to add a flag that allows missing keys?
        new = from_state_dict(value, state_dict, with_prefix(prefix, key))
        # Do not try to update parameters that are never defined
        if value is None and new is None:
            continue
        names.append(field.name)
        values.append(new)
    return eqx.tree_at(lambda m: [getattr(m, name) for name in names], mod, values)


def default_eqx_module_to_state_dict(mod: eqx.Module, prefix: Optional[str] = None) -> StateDict:
    """
    Convert an eqx.Module to a state dict. This is the default implementation of the to_state_dict method for
    eqx.Modules. It works by iterating over the fields of the module and calling to_state_dict on each field.
    Args:
        mod:
        prefix:

    Returns:

    """
    state_dict: StateDict = {}
    key_map: Dict[str, Optional[str]] = getattr(mod, "_state_dict_key_map", lambda: {})()  # type: ignore
    for field in dataclasses.fields(mod):
        if field.metadata.get("static", False):
            continue
        key = key_map.get(field.name, field.name)
        value = getattr(mod, field.name)
        child = to_state_dict(value, with_prefix(prefix, key))
        # TODO: should we check for conflicts?
        state_dict.update(child)
    return state_dict


def format_path_for_state_dict(prefix: Optional[str], path: Sequence) -> str:
    res = "".join(_format_key_path_element(path_elem) for path_elem in path)
    # res will have a .
    if prefix is not None:
        res = f"{prefix}{res}"
    elif res.startswith("."):
        res = res[1:]

    return res


# Torch compatible KeyPath formatting. Torch just always uses .
def _format_key_path_element(path_elem) -> str:
    match path_elem:
        case SequenceKey(idx):  # type: ignore
            return f".{idx}"
        case DictKey(key):  # type: ignore
            return f".{key}"
        case GetAttrKey():  # type: ignore
            return str(path_elem)
        case FlattenedIndexKey(idx):  # type: ignore
            return f".{idx}"
        case _:
            # The convention in JAX is to append the separator in the element itself
            # so we expect it to have
            path_elem = str(path_elem)
            if path_elem.startswith("."):
                return path_elem
            else:
                return f".{path_elem}"


def to_numpy_state_dict(model, prefix: Optional[str] = None) -> StateDict:
    """
    Convert a model to a state dict by first creating desharded copies of all parameters that reside in CPU
    memory.

    This method is especially useful for saving models distributed across multiple hosts.
    """

    with jax.default_device(jax.local_devices(backend="cpu")[0]):

        def get_to_cpu(arr):
            if is_scalarish(arr):
                return arr
            elif not is_jax_array_like(arr):
                return arr
            elif isinstance(arr, np.ndarray):
                return arr
            elif is_jax_array_like(arr):
                if arr.is_fully_addressable:
                    r = np.array(arr)
                    return r
                else:
                    # unfortunately, jax's allgather seems to replicate to every device rather than every host
                    # which doesn't work for ~7B parameter models on TPU (assuming we also have optimizer state)
                    # this approach limits us to <64B parameters, but that's good enough for now
                    # we're going to do something a bit fancy, where we shard the model into a (process, device) mesh,
                    # then look for some axis along which we can shard the array, and then we'll do an allgather
                    # via pjit. If we can't find one, we'll just fully replicate since it probably isn't that big.
                    # TODO: ensure that this mesh arranges devices correctly
                    # (jax seems to do this internally itself, so we should be fine?)
                    process_mesh = Mesh(
                        np.array(jax.devices()).reshape((jax.process_count(), -1)), ("process", "device")
                    )

                    # now we need to find an axis along which we can shard the array.
                    # for this, we need to find an axis s.t. size(axis) % local_devices == 0

                    try:
                        axis_to_shard = index_where(
                            lambda axis_size: axis_size % process_mesh.devices.size == 0, arr.shape
                        )
                    except ValueError:
                        return np.array(arr)

                    shardings = [None if i != axis_to_shard else "device" for i in range(len(arr.shape))]
                    sharding = NamedSharding(process_mesh, PartitionSpec(*shardings))
                    out = jax.device_put(arr, sharding)
                    return np.array(out)
            elif is_scalarish(arr):
                return np.asarray(arr)
            else:
                raise ValueError(f"Unsupported type {type(arr)}")

        # need to make sure the model is on *this machine* and *this machine's CPU* before saving
        model = jax.tree.map(lambda arr: get_to_cpu(arr), model)
        # TODO: it would be nice if safetensors supported an iterator or something so we could do the allgather one at a time
        state_dict = to_state_dict(model, prefix=prefix)
        return state_dict


_GLOBAL_SAVE_COUNT = 0


def save_state_dict(state_dict: StateDict, path):
    """
    Save a model's state dict to a file, bringing all tensors to the CPU first and then converting to numpy.
    This will save using safetensors format
    """
    state_dict = {k: v for k, v in state_dict.items() if v is not None}
    if jax.process_index() == 0:
        # the "pt" is a lie but it doesn't seem to actually matter and HF demands it
        safetensors.numpy.save_file(state_dict, path, metadata={"format": "pt"})
    global _GLOBAL_SAVE_COUNT
    sync_global_devices(f"save_state_dict {_GLOBAL_SAVE_COUNT}")
    _GLOBAL_SAVE_COUNT += 1


def load_state_dict(path):
    """
    Load a model's state dict from a file, bringing all tensors to the CPU first and then converting to numpy.
    This will load using safetensors format
    """
    state_dict = safetensors.numpy.load_file(path)
    return state_dict



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


from __future__ import annotations

from typing import Any, Callable

import equinox as eqx


def field(
    *,
    converter: Callable[[Any], Any] | None = None,
    static: bool = False,
    axis_names: tuple[str, ...] | None = None,
    **kwargs,
):
    """Wrapper around :func:`equinox.field` with optional ``axis_names`` metadata.

    Args:
        converter: Optional function applied to the value during dataclass initialisation.
        static: Whether the field is static in the PyTree.
        axis_names: Optional axis names associated with array fields. Cannot be
            specified together with ``static=True``.
        **kwargs: Additional keyword arguments forwarded to :func:`dataclasses.field`.

    Returns:
        A dataclasses field configured like :func:`equinox.field` with additional
        ``axis_names`` metadata.
    """
    if static and axis_names is not None:
        raise ValueError("axis_names cannot be specified together with static=True")

    metadata = dict(kwargs.pop("metadata", {}))
    metadata["axis_names"] = axis_names

    field_kwargs = {}
    if converter is not None:
        field_kwargs["converter"] = converter

    return eqx.field(static=static, metadata=metadata, **field_kwargs, **kwargs)



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


import dataclasses
import functools as ft
import inspect
from typing import Any, Callable, Literal, ParamSpec, Protocol, Sequence, Tuple, TypeVar, Union, overload

import equinox as eqx
import jax
import jax.tree_util as jtu
from jaxtyping import PyTree

import haliax
import haliax.tree_util as htu
from haliax._src.util import index_where
from haliax.axis import Axis, AxisSelector, selects_axis
from haliax.core import NamedArray
from haliax.jax_utils import is_jax_array_like, multilevel_scan, tree_checkpoint_name
from haliax.util import is_jax_or_hax_array_like, is_named_array

BoolAxisSpec = Union[bool, Callable[[Any], bool]]
Carry = TypeVar("Carry")
X = TypeVar("X", contravariant=True)
Y = TypeVar("Y", covariant=True)
Args = ParamSpec("Args")


def is_named_or_shaped_array_like(x):
    return (is_jax_array_like(x) and x.ndim >= 1) or is_named_array(x)


class ScanFn(Protocol[Carry, Args, Y]):
    """ """

    def __call__(self, carry: Carry, *args: Args.args, **kwargs: Args.kwargs) -> tuple[Carry, Y]: ...


@dataclasses.dataclass(frozen=True)
class ScanCheckpointPolicy:
    """
    A class that represents a gradient checkpoint policy for blocks in a Stacked module. This is used to control
    gradient checkpointing in [haliax.scan][], [haliax.fold][], [haliax.nn.Stacked][] and [haliax.nn.BlockSeq][].

    Gradient checkpointing is a technique for reducing memory usage in training large models. It works by saving only a
    subset of the forward pass and recomputing the rest in the backward pass. (By doing parts of the forward pass again)
    JAX suggests that this usually isn't necessary when not using scan-over-layers (i.e. Stacked), so this is mostly
    useful for Stacked modules.

    A scan block takes a "carry" and some extra arguments, and returns a "carry" and an "output". The "carry" is passed
    to the next block, and the "output" is concatenated into a final result (sort of like an RNN).

    Schematically it might look like this:

    ```
          I       I       I       I
          |       |       |       |
    C ->  B -C->  B -C->  B -C->  B --> C
          |       |       |       |
          O       O       O       O
    ```

    where "C" is the carry and "O" is the output. A block will typically do some computation (e.g. a Transformer block)
    as well, which might require saving or recomputing in the backward pass.

    Without checkpointing, we will save all block inputs and outputs, and use them to compute the gradient. This requires
    memory, of course. With checkpointing, we can save only some of the computation. Typically, to compute the gradient
    of the scan, we need the inputs to each block, the intermediates within each block, as well as the carries (which
    are inputs to the next block). We can save all of these, or we can save only some of them.

    With this class, you can specify if you want to save the carries, or the internals/inputs. You can also offload
    the carries to the host, which can reduce memory usage on the device.

    #### Nested Remat

    Alternatively, we can do something a bit more clever. We can break the computation into "blocks" of size "B", and
    save the carries and outputs of each block. Then, during the backward pass, we can recompute the outputs of each
    block using the carries and inputs, and then compute the gradient as usual. This requires O(B) memory and O(N)
    time. When B = sqrt(N), this is O(sqrt(N)) memory and O(N) time. This is the "nested scan" policy.
    In practice, this is about 20% slower than the O(N) memory policy, but it can be worth it for large models.

    #### Offloading

    Another choice is to "offload" carries and outputs to the host, which can reduce memory usage on the device.
    We support offloading carries and outputs to the host, but not internals.

    See Also:
        * [JAX docs on gradient checkpointing](https://docs.jax.dev/en/latest/gradient-checkpointing.html)
    """

    save_carries: bool | Literal["offload"] = True
    """
    Whether to save all carries in the forward pass. If True, (input) carries are saved in the forward pass and used in the
    backward pass. If "offload", carries are saved in the forward pass and offloaded to the host

    If False, carries are recomputed in the backward pass.
    """

    save_inputs: bool | Literal["offload"] = False
    """
    Whether to save all non-carry inputs in the forward pass. If True, inputs are saved in the forward pass and used in
    the backward pass. If "offload", inputs are saved in the forward pass and offloaded to the host.
    """

    save_block_internals: bool | list[str] = False
    """
    Whether to save internal state of blocks. If a list, only the listed names are saved, as
    with [jax.checkpoint_policies.save_only_these_names][].

    See Also: https://docs.jax.dev/en/latest/gradient-checkpointing.html#custom-policies-for-offload
    """

    offload_block_internals: list[str] = dataclasses.field(default_factory=list)
    """
    List of named block internals to offload to the host. This is useful for reducing memory usage on the device
    while still avoiding rematerialization.
    """

    prevent_cse: bool = False
    """
    Whether to prevent common subexpression elimination in the checkpointed function.
    """

    disable: bool = False
    """
    Whether to disable gradient checkpointing entirely. This is useful for debugging.
    """

    simple: bool = False
    """
    Whether to use the simple gradient checkpointing policy. This is useful for debugging.
    """

    nested: bool | int = False
    """
    Allows for nested remat with a double scan. We reshape the stack into [nested_remat, -1] and then scan over both
    in sequence. If True, we find the closest int to sqrt(len(stack)) such that len(stack) % int == 0.
    If False, we don't do anything.
    """

    @staticmethod
    def from_bool_or_str(remat_policy: bool | str):
        """
        Convert a boolean or string into a BlockCheckpointPolicy. This is useful for converting user input
        into a BlockCheckpointPolicy.

        Choices:
            * True: save outputs, don't save block internals. This is the classic Haliax behavior.
            * False: save everything.
            * "offload": offload outputs to the host, don't save block internals.
            * "recompute" or "full": don't save outputs or block internals.
            * "save_all": save outputs and block internals. Equivalent to False
            * "nested": use nested remat. Equivalent to True
        """
        if remat_policy == "offload":
            return ScanCheckpointPolicy(save_carries="offload", save_inputs="offload", save_block_internals=False)
        elif remat_policy == "recompute" or remat_policy == "full":
            return ScanCheckpointPolicy(save_carries=False, save_inputs=False, save_block_internals=False)
        elif remat_policy == "save_all":
            return ScanCheckpointPolicy(save_carries=True, save_inputs=True, save_block_internals=True)
        elif remat_policy == "nested":
            return ScanCheckpointPolicy(nested=True)
        elif remat_policy is True:
            return ScanCheckpointPolicy(simple=True)
        elif remat_policy is False:
            return ScanCheckpointPolicy(save_carries=True, save_inputs=True, save_block_internals=True, disable=True)
        else:
            raise ValueError(f"Invalid checkpoint policy {remat_policy}")

    @staticmethod
    def _mk(remat_policy: Union[bool, str, "ScanCheckpointPolicy"]) -> "ScanCheckpointPolicy":
        if isinstance(remat_policy, ScanCheckpointPolicy):
            return remat_policy
        else:
            return ScanCheckpointPolicy.from_bool_or_str(remat_policy)

    def checkpoint(self, carry_name: str, input_name: str, callable):
        if self.disable:
            return callable
        elif self.simple:
            return eqx.filter_checkpoint(callable, prevent_cse=self.prevent_cse)
        else:
            policy = self._to_jax_policy(carry_name, input_name)
            return eqx.filter_checkpoint(callable, policy=policy, prevent_cse=self.prevent_cse)

    def _to_jax_policy(self, carry_name: str, input_name: str):
        assert isinstance(carry_name, str)
        assert isinstance(input_name, str)
        our_names_to_save = []
        our_names_to_offload = []
        our_names_to_remat = []

        # return jax.checkpoint_policies.save_only_these_names(carry_name, output_name)

        if self.save_inputs is True:
            our_names_to_save.append(input_name)
        elif self.save_inputs == "offload":
            our_names_to_offload.append(input_name)
        else:
            assert self.save_inputs is False, f"Invalid save_inputs {self.save_inputs}"
            our_names_to_remat.append(input_name)

        if self.save_carries is True:
            our_names_to_save.append(carry_name)
        elif self.save_carries == "offload":
            our_names_to_offload.append(carry_name)
        else:
            assert self.save_carries is False, f"Invalid save_carries {self.save_carries}"
            our_names_to_remat.append(carry_name)

        if isinstance(self.save_block_internals, Sequence):
            our_names_to_save.extend(self.save_block_internals)

        if self.offload_block_internals:
            our_names_to_offload.extend(self.offload_block_internals)

        if not our_names_to_save and not our_names_to_offload and not self.save_block_internals:
            return None

        if len(our_names_to_offload) > 0:
            if self.save_block_internals is True:
                raise ValueError("Can't save all block internals and offload too. Use a list of names instead.")

            return jax.checkpoint_policies.save_and_offload_only_these_names(
                names_which_can_be_saved=our_names_to_save,
                names_which_can_be_offloaded=our_names_to_offload,
                offload_src="device",
                offload_dst="pinned_host",
            )
        else:
            if len(our_names_to_remat) > 0:
                if self.save_block_internals is True:
                    p1 = jax.checkpoint_policies.save_anything_except_these_names(*our_names_to_remat)
                    if len(our_names_to_save) > 0:
                        p2 = jax.checkpoint_policies.save_only_these_names(*our_names_to_save)
                        return jax.checkpoint_policies.save_from_both_policies(p1, p2)
                    else:
                        return p1
                else:
                    return jax.checkpoint_policies.save_only_these_names(*our_names_to_save)
            elif len(our_names_to_save) > 0:
                p1 = jax.checkpoint_policies.save_only_these_names(*our_names_to_save)
                if self.save_block_internals is True:
                    p2 = jax.checkpoint_policies.save_anything_except_these_names(*our_names_to_remat)
                    return jax.checkpoint_policies.save_from_both_policies(p1, p2)
                else:
                    return p1
            elif self.save_block_internals is True:
                return jax.checkpoint_policies.save_anything_except_these_names(*our_names_to_remat)
            else:
                return None


ScanCheckpointSpec = Union[ScanCheckpointPolicy, bool, Literal["offload", "recompute", "full", "save_all", "nested"]]


@overload
def scan(
    f: Callable[[Carry, X], tuple[Carry, Y]],
    axis: AxisSelector,
    *,
    remat: ScanCheckpointSpec = False,
    reverse: bool = False,
    unroll: int = 1,
    is_scanned: BoolAxisSpec = is_named_or_shaped_array_like,
) -> Callable[[Carry, PyTree[X]], tuple[Carry, PyTree[Y]]]: ...


@overload
def scan(
    f: Callable,
    axis: AxisSelector,
    *,
    remat: ScanCheckpointSpec = False,
    reverse: bool = False,
    unroll: int = 1,
    is_scanned: BoolAxisSpec = is_named_or_shaped_array_like,
) -> Callable: ...


def scan(
    f: Callable,  # : ScanFn[Carry, Args, Y],  This confuses mypy too much
    axis: AxisSelector,
    *,
    remat: ScanCheckpointSpec = False,
    reverse=False,
    unroll=1,
    is_scanned: BoolAxisSpec = is_named_or_shaped_array_like,
):
    """
    Scan over a named axis. Non-scalar unnamed arrays will have their first axis scanned over.

    Unlike [jax.lax.scan][], this function is curried: it takes the function, axis, and configuration arguments first, and
    then the initial carry and then any arguments to scan over as a separate curried function call.

    That is, `scan(f, axis)(init, xs)` is equivalent to `jax.lax.scan(f, init, xs)`

    Args:
        f (Callable): function to scan over
        axis (AxisSelector): axis to scan over
        reverse (bool): if True, scan in reverse
        unroll (int): unroll the loop by this amount
        remat: a ScanCheckpointPolicy or a boolean. If True, rematerialize block internals during
            gradient computation. If False, rematerialize nothing. If "offload", offload all carries and inputs.
            See [haliax.ScanCheckpointPolicy][] for more information.
        is_scanned (BoolAxisSpec): a function that takes a leaf of the tree and returns True if it should be scanned
            over, False otherwise. Behaves similarly to the `default` argument in filter_jit
    """
    checkpoint = ScanCheckpointPolicy._mk(remat)

    if isinstance(is_scanned, bool):
        q = is_scanned  # this is to make mypy happy
        is_scanned = lambda _: q

    def is_scanned_with_axis(leaf):
        if is_named_array(leaf):
            return selects_axis(leaf.axes, axis) and is_scanned(leaf)
        else:
            return is_scanned(leaf)

    def scanned_f(init, *args, **kwargs):
        # This implementation is a bit tricky.

        # first we want to partition the arguments into scanned and unscanned
        # unscanned arguments are just passed through, essentially captured as part of a lambda
        # scanned arguments are passed through the scan, which means we need to hoist the axis to the front
        xs = (args, kwargs)
        scanned_xs, unscanned_xs = eqx.partition(xs, is_scanned_with_axis, is_leaf=is_named_array)

        carry_name = f"scan({haliax.axis_name(axis)})__carry"
        input_name = f"scan({haliax.axis_name(axis)})__input"

        # Next we have to hoist the axis we're scanning over to the front of the array, because that's what scan
        # expects. Then we have to scan over the 0th dim of the arrays (as flattened non-pytrees)
        # We have to be careful that we don't try to create NamedArrays that have the shape of the scanned result
        # but don't yet have the scanned axis as ones of `axes`, so we use _ScannedArrayResult that doesn't check
        # invariants until we're ready to create the result.
        axis_first_xs = htu.tree_map(_ensure_first(axis), scanned_xs)

        # now get a template of an element of "X"
        x_elem = htu.tree_map(_select_0th(axis), axis_first_xs)
        # NB: we don't want to use htu.tree_structure here because we want to eliminate the leading axis
        x_elem_structure = jax.tree_util.tree_structure(x_elem)

        # now we can fold over the axis
        @ft.wraps(f)
        def wrapped_fn(carry, scanned_x_leaves):
            scanned_x = jax.tree_util.tree_unflatten(x_elem_structure, scanned_x_leaves)
            scanned_x = tree_checkpoint_name(scanned_x, input_name)
            carry = tree_checkpoint_name(carry, carry_name)

            # this part is the most delicate: combining the scanned x with the unscanned x
            scanned_x = eqx.combine(scanned_x, unscanned_xs, is_leaf=is_named_array)
            args, kwargs = scanned_x
            carry, y = f(carry, *args, **kwargs)
            y = htu.tree_map(_pacify_named_arrays, y)

            return carry, y

        true_axis = _infer_axis_size_from_tree(axis_first_xs, axis)
        axis_size = true_axis.size

        # build a mapping from positional argument indices to their names for friendlier error messages
        sig = inspect.signature(f)
        arg_pos_names: dict[int, str] = {}
        params = list(sig.parameters.values())[1:]  # skip carry
        pos_count = 0
        var_pos_name: str | None = None
        for param in params:
            if param.kind in (
                inspect.Parameter.POSITIONAL_ONLY,
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
            ):
                arg_pos_names[pos_count] = param.name
                pos_count += 1
            elif param.kind == inspect.Parameter.VAR_POSITIONAL:
                var_pos_name = param.name
                break
        if var_pos_name is not None:
            for i in range(pos_count, len(args)):
                arg_pos_names[i] = f"{var_pos_name}[{i - pos_count}]"

        path_leaves, _ = jtu.tree_flatten_with_path(axis_first_xs, is_leaf=is_named_array)
        mismatched = []
        for path, leaf in path_leaves:
            if isinstance(leaf, NamedArray):
                lead_size = leaf.array.shape[0]
            elif is_jax_array_like(leaf):
                lead_size = leaf.shape[0]
            else:
                continue
            if lead_size != axis_size:
                mismatched.append((path, lead_size))
        if mismatched:
            details = ", ".join(
                f"{_format_tree_path(p, arg_pos_names)} has leading dimension {s}" for p, s in mismatched
            )
            raise ValueError(
                f"scan got `length` argument of {axis_size} but some inputs had different leading axis sizes: {details}"
            )

        nested_scan = checkpoint.nested
        outer_block_size = nested_scan_outer_block(nested_scan, axis_size)

        checkpointed_fn = checkpoint.checkpoint(carry_name, input_name, wrapped_fn)

        # as above, we don't want to use htu.tree_leaves here because we want to eliminate the leading axis
        leaves = jax.tree_util.tree_leaves(axis_first_xs)

        with jax.named_scope(f"scan({haliax.axis_name(axis)})"):
            if outer_block_size is not None:
                carry, ys = multilevel_scan(
                    checkpointed_fn, init, leaves, outer_block_size, reverse=reverse, unroll=unroll, length=axis_size
                )
            else:
                carry, ys = jax.lax.scan(
                    checkpointed_fn, init, leaves, reverse=reverse, unroll=unroll, length=axis_size
                )

        ys = jax.tree_util.tree_map(_prepend_named_batch_axis(true_axis), ys, is_leaf=_is_passive_array)

        return carry, ys

    def nested_scan_outer_block(nested_remat, axis_size):
        outer_block_size: int | None
        if nested_remat is True:
            outer_block_size = find_closest_divisible_int_to_sqrt(axis_size)
        elif nested_remat is False:
            outer_block_size = None
        else:
            outer_block_size = nested_remat
        return outer_block_size

    return scanned_f


@overload
def fold(
    fn: Callable[[Carry, X], Carry],
    axis: AxisSelector,
    *,
    remat: ScanCheckpointSpec = False,
    reverse: bool = False,
    unroll: int = 1,
    is_scanned: BoolAxisSpec = is_named_or_shaped_array_like,
) -> Callable[[Carry, PyTree[X]], Carry]: ...


@overload
def fold(
    fn: Callable,
    axis: AxisSelector,
    *,
    remat: ScanCheckpointSpec = False,
    reverse: bool = False,
    unroll: int = 1,
    is_scanned: BoolAxisSpec = is_named_or_shaped_array_like,
) -> Callable: ...


def fold(
    fn: Callable,
    axis: AxisSelector,
    *,
    remat: ScanCheckpointSpec = False,
    reverse: bool = False,
    unroll: int = 1,
    is_scanned: BoolAxisSpec = is_named_or_shaped_array_like,
) -> Callable:
    """
    Slightly simpler implementation of scan that folds over the named axis of the array, not returning intermediates.

    As with scan, this function is curried: it takes the function, axis, and configuration arguments first, and
    then the initial carry and then any arguments to scan over as a separate curried function call.

    Unnamed arrays will have their first axis scanned over, unless they are scalars, in which case they will be passed
    through unchanged.

    Args:
        fn: function to reduce over
        axis: axis to reduce over
        reverse: if True, reduce in reverse
        unroll: unroll the loop by this amount
        nested_scan: Use nested scans to reduce memory usage. If an integer, use that many nested scans.
                If true, use the closest int to sqrt(axis.size) that divides axis.size, which gives
                O(sqrt(N)) memory and O(N) time.
        is_scanned: a function that takes a leaf of the tree and returns True if it should be scanned over,
                    False otherwise. Behaves similarly to the `default` argument in filter_jit

    Returns:
        A function that takes the initial carry and then the arguments to reduce over, and returns the final carry
    """

    def scan_compatible_fn(carry, *args, **kwargs):
        return fn(carry, *args, **kwargs), None

    scan_preconfig = scan(scan_compatible_fn, axis, reverse=reverse, unroll=unroll, is_scanned=is_scanned, remat=remat)

    def scanned_f(init, *args, **kwargs):
        return scan_preconfig(init, *args, **kwargs)[0]

    return scanned_f


def map(
    fn: Callable[[X], Y],
    axis: Axis,
    *,
    remat: ScanCheckpointSpec = False,
    reverse: bool = False,
    unroll: int = 1,
    is_mapped: BoolAxisSpec = is_jax_or_hax_array_like,
) -> Callable[[PyTree[X]], PyTree[Y]]:
    """
    NamedArray aware version of jax.lax.map. Normal arrays are mapped according to the specs as in equinox.filter_map,
    except that the output axis is always 0 b/c it's annoying to make anything else work.

    You'll typically want to use map (instead of a vmap or just vectorized code) when you want to encourage XLA to
    loop over the axis to control memory.

    Args:
        fn (Callable):  function to map over
        axis (Axis): axis to map over
        reverse (bool): if True, map in reverse
        unroll (int): unroll the loop by this amount

    """

    def scan_compatible_fn(_, x):
        del _
        return None, fn(x)

    scan_preconfig = scan(scan_compatible_fn, axis, reverse=reverse, unroll=unroll, is_scanned=is_mapped, remat=remat)

    def scanned_f(*args, **kwargs):
        return scan_preconfig(None, *args, **kwargs)[1]

    return scanned_f


ResolvedUnnamedAxisSpec = Union[int, None]
UnnamedAxisSpec = Union[ResolvedUnnamedAxisSpec, Callable[[Any], ResolvedUnnamedAxisSpec]]


def _zero_if_array_else_none(x: Any) -> ResolvedUnnamedAxisSpec:
    return 0 if is_jax_array_like(x) else None


def _format_tree_path(path: tuple[jtu.KeyEntry, ...], arg_pos_names: dict[int, str] | None = None) -> str:
    parts: list[str] = []
    i = 0
    if len(path) >= 2 and isinstance(path[0], jtu.SequenceKey):
        if path[0].idx == 0 and isinstance(path[1], jtu.SequenceKey):
            name = (arg_pos_names or {}).get(path[1].idx)
            if name is not None:
                parts.append(name)
            else:
                parts.append(f"[{path[1].idx}]")
            i = 2
        elif path[0].idx == 1 and isinstance(path[1], jtu.DictKey):
            parts.append(str(path[1].key))
            i = 2
    for p in path[i:]:
        if isinstance(p, jtu.GetAttrKey):
            parts.append("." + p.name)
        elif isinstance(p, jtu.DictKey):
            parts.append(f"[{p.key!r}]")
        elif isinstance(p, jtu.SequenceKey):
            parts.append(f"[{p.idx}]")
        else:  # pragma: no cover - future-proofing
            parts.append(str(p))
    if parts and parts[0].startswith("."):
        parts[0] = parts[0][1:]
    return "".join(parts) or "<root>"


def _infer_axis_size_from_tree(result, axis):
    if isinstance(axis, str):
        result_leaves = jax.tree_util.tree_leaves(result, is_leaf=_is_passive_array)
        if len(result_leaves) == 0:
            # this really shouldn't happen
            return None
        if isinstance(result_leaves[0], _PassiveNamedArray):
            true_axis_size = result_leaves[0].array.shape[0]  # batch axis is defined to be 0 above
            true_axis = Axis(axis, true_axis_size)
        else:
            true_axis_size = result_leaves[0].shape[0]  # batch axis is defined to be 0 above
            true_axis = Axis(axis, true_axis_size)
    else:
        true_axis = axis
    return true_axis


@jax.tree_util.register_pytree_node_class
@dataclasses.dataclass(frozen=True)
class _PassiveNamedArray:
    """For some higher-order jax manipulations, jax will insert/remove an axis at the beginning of the array (or
    sometimes elsewhere). Jax doesn't know about our NamedArray wrapper, so we need a variant of NamedArray
    that doesn't care about the axis names. This is that variant.This class is a 'chill' version of NamedArray
    that doesn't check invariants until we're ready to create the result

    For example, with scan in NamedArray, we can't just have the scan tree prepend the scanned axis to the result,
    because we don't have a way to feed it the name of the scanned axis.
    """

    array: jax.numpy.ndarray
    main_axes: Tuple[Axis, ...]

    def as_scanned_result(self, scan_axis: Axis):
        return NamedArray(self.array, (scan_axis,) + self.main_axes)

    def strip_axis(self, axis: AxisSelector):
        if isinstance(axis, Axis):
            index = self.main_axes.index(axis)
        else:
            index = index_where(lambda a: a.name == axis, self.main_axes)
        return NamedArray(self.array, self.main_axes[:index] + self.main_axes[index + 1 :])

    def to_named_array(self):
        return NamedArray(self.array, self.main_axes)

    def tree_flatten(self) -> Any:
        return ((self.array,), self.main_axes)

    @classmethod
    def tree_unflatten(cls, aux, tree: Any) -> Any:
        assert len(tree) == 1
        return cls(tree[0], main_axes=aux)


def _pacify_named_arrays(leaf):
    if isinstance(leaf, NamedArray):
        return _PassiveNamedArray(leaf.array, leaf.axes)
    elif isinstance(leaf, _PassiveNamedArray):
        assert False, "PassiveNamedArray should not be present in the tree"
    else:
        return leaf


def _select_0th(axis):
    def select_0th(leaf):
        if isinstance(leaf, NamedArray):
            return leaf.take(axis, 0)
        elif isinstance(leaf, _PassiveNamedArray):
            assert False, "PassiveNamedArray should not be present in the tree"
        else:
            # other leaves don't matter
            return leaf

    return select_0th


def _ensure_first(axis):
    def ensure_first(leaf):
        if isinstance(leaf, NamedArray):
            return leaf.rearrange((axis, ...))
        elif isinstance(leaf, _PassiveNamedArray):
            assert False, "PassiveNamedArray should not be present in the tree"
        else:
            return leaf

    return ensure_first


def find_closest_divisible_int_to_sqrt(n: int) -> int:
    """
    Find the closest integer to the square root of n (less than or equal to sqrt(n)) that divides n.
    """
    assert n > 0, f"Expected n > 0, got {n}"
    for i in range(int(n**0.5), 0, -1):
        if n % i == 0:
            return i

    return 1


def _is_passive_array(arr):
    return isinstance(arr, _PassiveNamedArray)


def _prepend_named_batch_axis(leading_axis: Axis):
    def to_active_named_array(leaf):
        if isinstance(leaf, _PassiveNamedArray):
            return leaf.as_scanned_result(leading_axis)
        else:
            return leaf

    return to_active_named_array



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


import equinox as eqx
import jax
import jax.numpy as jnp

import haliax
import haliax as hax
from haliax.nn.pool import max_pool, mean_pool

# Tests largely cribbed from equinox


def test_maxpool1d():
    D = hax.Axis("D", 14)
    x = hax.arange(D)
    output = max_pool((D.resize(2),), x, stride=(3,))
    answer = jnp.array([1, 4, 7, 10, 13], dtype=jnp.int32)

    assert jnp.all(output.array == answer)

    answer = jnp.array([2, 5, 8, 11])
    output = max_pool(D.resize(3), x, stride=(3,), padding=0)
    assert jnp.all(output.array == answer)

    # max_pool = eqx.nn.MaxPool1d(kernel_size=3, stride=3, padding=0, use_ceil=True)
    answer = jnp.array([2, 5, 8, 11, 13])
    output = max_pool(D.resize(3), x, stride=(3,), padding=0, use_ceil=True)
    assert jnp.all(output.array == answer)

    # test batch axes
    B = hax.Axis("B", 2)
    x = x.rearrange("(B D) -> B D", B=B)
    output = max_pool(D.resize(2), x, stride=(3,), padding="VALID")
    answer = jnp.array([[1, 4], [8, 11]])
    assert jnp.all(output.array == answer)

    output = max_pool(D.resize(2), x, stride=(3,), use_ceil=True)
    answer = jnp.array([[1, 4, 6], [8, 11, 13]])
    assert jnp.all(output.array == answer)

    output = max_pool(D.resize(3), x, stride=(3,), padding=0)
    answer = jnp.array([[2, 5], [9, 12]])
    assert jnp.all(output.array == answer)

    output = max_pool(D.resize(3), x, stride=(3,), padding=0, use_ceil=True)
    answer = jnp.array([[2, 5, 6], [9, 12, 13]])
    assert jnp.all(output.array == answer)

    output = max_pool(D.resize(2), x, stride=(3,), padding="SAME")
    answer = jnp.array([[1, 4, 6], [8, 11, 13]])
    assert jnp.all(output.array == answer)


def test_maxpool2d():
    _x = jnp.arange(36).reshape(6, 6)
    x = hax.named(_x, ("H", "W"))

    # max_pool = eqx.nn.MaxPool2d(2, (3, 2))
    output = max_pool((hax.Axis("H", 2), hax.Axis("W", 2)), x, stride=(3, 2))
    answer = jnp.array([[7, 9, 11], [25, 27, 29]])

    assert jnp.all(output.array == answer)

    output = max_pool((hax.Axis("H", 3), hax.Axis("W", 3)), x, stride=2, padding=1)
    answer = jnp.array([[7, 9, 11], [19, 21, 23], [31, 33, 35]])

    assert jnp.all(output.array == answer)

    # test batch axes
    B = hax.Axis("B", 2)
    x = haliax.stack(B, [x, x])

    output = max_pool((hax.Axis("H", 2), hax.Axis("W", 2)), x, stride=(3, 2))
    answer = jnp.array([[[7, 9, 11], [25, 27, 29]], [[7, 9, 11], [25, 27, 29]]])

    assert jnp.all(output.array == answer)


def test_maxpool3d():
    _x = jnp.arange(64).reshape(4, 4, 4)
    x = hax.named(_x, ("H", "W", "D"))
    output = max_pool((hax.Axis("H", 2), hax.Axis("W", 2), hax.Axis("D", 2)), x, stride=(3, 2, 1))

    answer = jnp.array([[[21, 22, 23], [29, 30, 31]]])

    assert jnp.all(output.array == answer)

    answer = jnp.asarray(
        [
            [[37, 39, 39], [45, 47, 47], [45, 47, 47]],
            [[53, 55, 55], [61, 63, 63], [61, 63, 63]],
        ]
    )
    output = max_pool(
        (hax.Axis("H", 3), hax.Axis("W", 3), hax.Axis("D", 3)),
        x,
        stride=2,
        padding=((0, 1), (1, 1), (1, 1)),
        use_ceil=True,
    )
    assert jnp.all(output.array == answer)


def test_mean_pool_1d():
    D = hax.Axis("D", 14)
    x = hax.arange(D)
    output = mean_pool((D.resize(2),), x, stride=(3,))
    answer = jnp.array([0.5, 3.5, 6.5, 9.5, 12.5])

    assert jnp.all(output.array == answer)

    # no pad
    output = mean_pool(D.resize(3), x, stride=(3,), padding=0)
    answer = jnp.array([1, 4, 7, 10])
    assert jnp.all(output.array == answer)

    # pad, no include pad in avg
    output = mean_pool(D.resize(3), x, stride=(3,), padding="SAME", count_include_pad=False)
    answer = jnp.array([1, 4, 7, 10, 12.5])

    assert jnp.all(output.array == answer)

    output = mean_pool(D.resize(3), x, stride=(3,), padding="SAME", count_include_pad=True)
    answer = jnp.array([1, 4, 7, 10, (12 + 13) / 3.0])

    assert jnp.all(output.array == answer)


def test_mean_pool_2d():
    _x = jnp.arange(36).reshape(6, 6)
    x = hax.named(_x, ("H", "W"))

    output = mean_pool((hax.Axis("H", 1), hax.Axis("W", 3)), x, stride=2)
    answer = jnp.array([[1, 3], [13, 15], [25, 27]])

    assert jnp.all(output.array == answer)

    # test batch axes
    B = hax.Axis("B", 2)
    x = haliax.stack(B, [x, x])

    output = mean_pool((hax.Axis("H", 1), hax.Axis("W", 3)), x, stride=2)
    answer = jnp.array([[[1, 3], [13, 15], [25, 27]], [[1, 3], [13, 15], [25, 27]]])

    assert jnp.all(output.array == answer)


def test_mean_pool3d():
    _x = jnp.arange(64).reshape(4, 4, 4)
    x = hax.named(_x, ("H", "W", "D"))
    output = mean_pool((hax.Axis("H", 1), hax.Axis("W", 3), hax.Axis("D", 1)), x, stride=2)

    answer = jnp.array([[[4, 6]], [[36, 38]]])

    assert jnp.all(output.array == answer)


def test_pool_backprop():
    def max_pool_mean(x):
        pooled = max_pool(
            (hax.Axis("H", 2), hax.Axis("W", 2), hax.Axis("D", 2)), x, stride=1, padding=((0, 1), (0, 1), (0, 1))
        )
        return hax.mean(pooled).scalar()

    _x = jnp.arange(64, dtype=jnp.float32).reshape(1, 4, 4, 4)
    x = hax.named(_x, ("B", "H", "W", "D"))
    grad_fn = jax.value_and_grad(max_pool_mean)

    hax_loss, hax_grad = grad_fn(x)

    # compare it to eqx

    eqx_max_pool = eqx.nn.MaxPool3d(2, (1, 1, 1), padding=((0, 1), (0, 1), (0, 1)))

    def eqx_max_pool_mean(x):
        pooled = eqx_max_pool(x)
        return pooled.mean()

    eqx_grad_fn = jax.value_and_grad(eqx_max_pool_mean)
    eqx_loss, eqx_grad = eqx_grad_fn(_x)

    assert jnp.allclose(hax_loss, eqx_loss)
    assert jnp.allclose(hax_grad.array, eqx_grad)



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


import dataclasses
from typing import List, Tuple, Union, Sequence

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.tree_util as jtu


from haliax.core import NamedArray
from haliax.axis import Axis
from haliax.util import is_jax_or_hax_array_like

from ._src.util import IdentityMap

ArrayLike = Union[jnp.ndarray, NamedArray]


def describe_array(arr):
    if isinstance(arr, NamedArray):
        return f"NamedArray(axes={arr.axes}, dtype={arr.dtype})"
    else:
        return f"ndarray(shape={arr.shape}, dtype={arr.dtype})"


class ModuleProblems(Exception):
    def __init__(self):
        self.reused_arrays: List[Tuple[ArrayLike, List]] = []
        self.static_arrays: List[str] = []

    def __bool__(self):
        return bool(self.reused_arrays or self.static_arrays)

    def __str__(self):
        if not self:
            return "No problems found"
        else:
            return "\n".join(
                [
                    "Found some problems with your module:",
                    *self._format_reused_arrays(),
                    *self._format_static_arrays(),
                ]
            )

    def _format_reused_arrays(self):
        return [f"  Reused array {describe_array(arr)} at paths {paths}" for arr, paths in self.reused_arrays]

    def _format_static_arrays(self):
        return [f"  Static array at field {field}" for field in self.static_arrays]


def diagnose_common_issues(module: eqx.Module):
    """
    Checks for common issues in a module, such as reused arrays and static arrays.
    Equinox modules (and therefore Haliax modules) should not have arrays that are stored
    in multiple places, and should not have arrays stored as static fields.

    We'll add more checks here as we find them.

    Args:
        module:  The module to check for problems

    Returns:
        None

    Raises:
        ModuleProblems: if any problems are found

    """

    problems = ModuleProblems()
    _check_for_reused_arrays(problems, module)
    _check_for_static_arrays(problems, module)

    if problems:
        raise problems

    # just in case we missed anything, raise equinox's errors:
    eqx.tree_check(module)


def _check_for_reused_arrays(problems, module):
    used_arrays = IdentityMap[ArrayLike, List[str]]()

    path_leaves, _ = jtu.tree_flatten_with_path(module, is_leaf=is_jax_or_hax_array_like)

    for path, leaf in path_leaves:
        if is_jax_or_hax_array_like(leaf):
            if leaf in used_arrays:
                used_arrays[leaf].append(jtu.keystr(path))
            else:
                used_arrays[leaf] = [jtu.keystr(path)]

    for arr, paths in used_arrays.items():
        if len(paths) > 1:
            problems.reused_arrays.append((arr, paths))


def _check_for_static_arrays(problems, module):
    static_arrays = []

    def recurse(module, path):
        if isinstance(module, eqx.Module):
            for field in dataclasses.fields(module):
                value = getattr(module, field.name)
                if field.metadata.get("static", False) and is_jax_or_hax_array_like(value):
                    static_arrays.append(f"{path}.{field.name}")
                else:
                    recurse(value, f"{path}.{field.name}")
        else:
            leaves, _ = eqx.tree_flatten_one_level(module)
            if leaves != [module]:
                leaves_with_names = jtu.tree_leaves_with_path(module, is_leaf=lambda x: x in leaves)
                for name, leaf in leaves_with_names:
                    recurse(leaf, f"{path}{name}")

    recurse(module, "")

    if static_arrays:
        problems.static_arrays.extend(static_arrays)


def _pspec_parts(spec_part) -> str:
    if spec_part is None:
        return "unsharded"
    elif isinstance(spec_part, (tuple, list)):
        return "+".join(str(p) for p in spec_part)
    else:
        return str(spec_part)


def visualize_named_sharding(axes: Sequence[Axis], sharding: jax.sharding.Sharding) -> None:
    """Visualize the sharding for a set of named axes.

    This extends :func:`jax.debug.visualize_sharding` to handle arrays with more
    than two dimensions by falling back to a textual description when necessary.
    """

    try:
        pspec = sharding.spec  # type: ignore[attr-defined]
    except Exception:
        pspec = (None,) * len(axes)

    parts = [_pspec_parts(p) for p in pspec]
    num_sharded = sum(p != "unsharded" for p in parts)

    if num_sharded <= 2:
        try:
            jax.debug.visualize_sharding([ax.size for ax in axes], sharding)
        except Exception:
            pass

    mapping = ", ".join(f"{ax.name}->{part}" for ax, part in zip(axes, parts))
    print(mapping)


def visualize_shardings(tree) -> None:
    """Print the sharding for each array-like leaf in ``tree``.

    Both :class:`NamedArray` and regular JAX arrays are supported. NamedArrays
    will show the mapping from logical axis names to physical axes. Plain arrays
    will fall back to :func:`jax.debug.visualize_sharding`.
    """

    import haliax.tree_util as htu

    def _show(x):
        if isinstance(x, NamedArray):
            arr = x.array
            axes = x.axes
        else:
            arr = x
            axes = None

        def cb(sh):
            if axes is not None:
                visualize_named_sharding(axes, sh)
            else:
                try:
                    jax.debug.visualize_sharding(arr.shape, sh)
                except Exception:
                    pass

        jax.debug.inspect_array_sharding(arr, callback=cb)
        return x

    htu.tree_map(_show, tree, is_leaf=is_jax_or_hax_array_like)



from typing import Any, Callable

import jax.numpy as jnp
import haliax as hax


def _sample_array():
    Height, Width = hax.make_axes(Height=2, Width=3)
    data = jnp.array([[1.0, jnp.nan, 3.0], [jnp.nan, 5.0, 6.0]])
    arr = hax.named(data, (Height, Width))
    return Height, Width, data, arr


def test_amin_alias():
    Height, Width = hax.make_axes(Height=2, Width=3)
    data = jnp.arange(6.0).reshape(2, 3)
    arr = hax.named(data, (Height, Width))
    assert jnp.array_equal(hax.amin(arr).array, jnp.amin(data))
    assert jnp.array_equal(hax.amin(arr, axis=Height).array, jnp.amin(data, axis=0))
    assert hax.amin(arr, axis=Height).axes == (Width,)


def test_nan_reductions():
    Height, Width, data, arr = _sample_array()

    funcs: list[tuple[Callable[..., Any], Callable[..., Any]]] = [
        (hax.nanmin, jnp.nanmin),
        (hax.nanmax, jnp.nanmax),
        (hax.nanmean, jnp.nanmean),
        (hax.nansum, jnp.nansum),
        (hax.nanprod, jnp.nanprod),
        (hax.nanstd, jnp.nanstd),
        (hax.nanvar, jnp.nanvar),
    ]

    for hfunc, jfunc in funcs:
        assert jnp.allclose(hfunc(arr).array, jfunc(data), equal_nan=True)
        assert jnp.allclose(hfunc(arr, axis=Height).array, jfunc(data, axis=0), equal_nan=True)
        assert hfunc(arr, axis=Height).axes == (Width,)

    arg_funcs: list[tuple[Callable[..., Any], Callable[..., Any]]] = [
        (hax.nanargmax, jnp.nanargmax),
        (hax.nanargmin, jnp.nanargmin),
    ]

    for hfunc, jfunc in arg_funcs:
        out = hfunc(arr, axis=Height)
        assert jnp.array_equal(out.array, jfunc(data, axis=0))
        assert out.axes == (Width,)

    axiswise_funcs: list[tuple[Callable[..., Any], Callable[..., Any]]] = [
        (hax.nancumsum, jnp.nancumsum),
        (hax.nancumprod, jnp.nancumprod),
    ]

    for hfunc, jfunc in axiswise_funcs:
        out = hfunc(arr, axis=Height)
        assert jnp.allclose(out.array, jfunc(data, axis=0), equal_nan=True)
        assert out.axes == (Height, Width)



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


from __future__ import annotations

import typing

import jax.numpy as jnp

from haliax import Axis, NamedArray
from haliax.haxtyping import Float, Int, f32, i32, Named


def test_namedarray_type_syntax():
    axes1 = typing.get_args(NamedArray["batch", "embed"])[1]
    axes2 = typing.get_args(Named[NamedArray, "batch embed"])[1]  # type: ignore
    assert axes1 == axes2

    axes3 = typing.get_args(NamedArray["batch embed ..."])[1]
    assert axes3.before == ("batch", "embed") and axes3.subset and axes3.after == ()

    axes4 = typing.get_args(NamedArray[{"batch", "embed"}])[1]
    assert set(axes4.before) == {"batch", "embed"} and not axes4.ordered

    axes5 = typing.get_args(NamedArray[{"batch", "embed", ...}])[1]
    assert set(axes5.before) == {"batch", "embed"} and not axes5.ordered and axes5.subset

    axes6 = typing.get_args(NamedArray["... embed"])[1]
    assert axes6.before == () and axes6.after == ("embed",) and axes6.subset

    axes7 = typing.get_args(NamedArray["batch ... embed"])[1]
    assert axes7.before == ("batch",) and axes7.after == ("embed",) and axes7.subset


def test_named_param_annotation():
    def foo(x: f32[NamedArray, "batch embed"]):  # type: ignore  # noqa: F722
        pass

    axes = typing.get_args(typing.get_type_hints(foo, include_extras=True)["x"])[1]
    assert axes.before == ("batch", "embed")


def test_namedarray_runtime_check():
    Batch = Axis("batch", 2)
    Embed = Axis("embed", 3)
    arr = NamedArray(jnp.zeros((Batch.size, Embed.size)), (Batch, Embed))
    assert arr.matches_axes(NamedArray["batch", "embed"])
    assert arr.matches_axes(Named[NamedArray, "batch embed"])  # type: ignore
    assert arr.matches_axes(NamedArray["batch embed ..."])
    assert arr.matches_axes(NamedArray[{"batch", "embed"}])
    assert arr.matches_axes(NamedArray[{"batch", "embed", ...}])
    assert not arr.matches_axes(NamedArray["embed batch"])
    assert not arr.matches_axes(NamedArray[{"batch", "foo", ...}])


def test_namedarray_runtime_check_with_dtype():
    Batch = Axis("batch", 2)
    arr = NamedArray(jnp.zeros((Batch.size,), dtype=jnp.float32), (Batch,))
    assert arr.matches_axes(f32["batch"])  # type: ignore
    assert not arr.matches_axes(i32["batch"])  # type: ignore


def test_namedarray_runtime_check_with_category():
    B = Axis("batch", 1)
    arr = NamedArray(jnp.zeros((B.size,), dtype=jnp.float32), (B,))
    assert arr.matches_axes(Float[NamedArray, "batch"])  # type: ignore
    assert not arr.matches_axes(Int[NamedArray, "batch"])  # type: ignore



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


import functools as ft
import typing
import warnings
from typing import Dict, Optional, Tuple

import jax

import haliax
from haliax.axis import (
    Axis,
    AxisSelection,
    PartialAxisSpec,
    axis_name,
    axis_spec_to_shape_dict,
    eliminate_axes,
    rearrange_for_partial_order,
    union_axes,
)
from haliax.core import NamedArray
from haliax.jax_utils import _jittable_dg_einsum
from haliax.types import DTypeLike, PrecisionLike


# deprecated overload
@typing.overload
def dot(
    axis: Optional[AxisSelection],
    *arrays: NamedArray,
    precision: PrecisionLike = None,
    preferred_element_type: Optional[DTypeLike] = None,
    out_axes: Optional[PartialAxisSpec] = ...,
    dot_general=jax.lax.dot_general,
) -> NamedArray: ...


@typing.overload
def dot(
    *arrays: NamedArray,
    axis: Optional[AxisSelection],
    precision: PrecisionLike = None,
    preferred_element_type: Optional[DTypeLike] = None,
    out_axes: Optional[PartialAxisSpec] = ...,
    dot_general=jax.lax.dot_general,
) -> NamedArray: ...


def dot(
    *arrays,
    precision: PrecisionLike = None,
    preferred_element_type: Optional[DTypeLike] = None,
    out_axes: Optional[PartialAxisSpec] = None,
    dot_general=jax.lax.dot_general,
    **kwargs,
) -> NamedArray:
    """Returns the tensor product of two NamedArrays. The axes `axis` are contracted over,
    and any other axes that are shared between the arrays are batched over. Non-contracted Axes in one
    that are not in the other are preserved.

    Note that if `axis` is None, the result will be a scalar, not a NamedArray. The semantics of `axis=None` are
    similar to e.g. how `sum` and other reduction functions work in numpy. If `axis=()`, then the result will be
    an "outer product" of the arrays, i.e. a tensor with shape equal to the concatenation of the shapes of the arrays.

    By default, the order of output axes is determined by the order of the input axes, such that each output axis
    occurs in the same order as it first occurs in the concatenation of the input axes.

    If `out_axes` is provided, the output will be transposed to match the provided axes. `out_axes` may be a partial
    specification of the output axes (using ellipses), in which case the output will be rearranged to be consistent
    with the partial specification. For example, if `out_axes=(..., Height, Width)` and the output axes are
    `(Width, Height, Depth)`, the output will be transposed to `(Depth, Height, Width)`. Multiple ellipses
    are supported, in which case axes will be inserted according to a greedy heuristic that prefers to place
    unconstrained axes as soon as all prior axes in the "natural" order are covered.

    Args:
        *arrays (NamedArray): The arrays to contract.
        axis (AxisSelection): The axes to contract over.
        precision (PrecisionLike, optional): The precision to use. Defaults to None. This argument is passed to `jax.numpy.einsum`,
            which in turn passes it to jax.lax.dot_general.
        preferred_element_type (DTypeLike, optional): The preferred element type of the result. Defaults to None.
            This argument is passed to `jax.numpy.einsum`.
        out_axes (Optional[PartialAxisSpec], optional): a potentially partial specification of the output axes.
            If provided, the output will be transposed to match the provided axes. Defaults to None.


    Returns:
        NamedArray: The result of the contraction.
    """
    if len(arrays) == 0:
        raise ValueError("Must provide at least one array to dot")

    if "axis" in kwargs:
        axis = kwargs["axis"]
    else:
        axis = arrays[0]
        arrays = arrays[1:]
        if isinstance(axis, NamedArray):
            raise ValueError("Must provide an axis to dot")

        warnings.warn("Axis has been changed to a keyword argument. Please update your code.", DeprecationWarning)

    _ensure_no_mismatched_axes(*arrays)

    # to call dot_general we need two things:
    # list of contractions and list of arrays

    all_axes: Tuple[Axis, ...] = ft.reduce(union_axes, (a.axes for a in arrays), ())  # type: ignore
    output_axes: Tuple[Axis, ...]
    if axis is None:
        # we want to contract over all the axes
        output_axes = ()
    else:
        output_axes = eliminate_axes(all_axes, axis)

    if out_axes is not None:
        output_axes = rearrange_for_partial_order(out_axes, output_axes)

    array_specs = []

    next_index = 0
    axis_mappings: Dict[str, int] = {}

    for a in arrays:
        spec = ""
        for ax in a.axes:
            if ax.name in axis_mappings:
                spec += f"{axis_mappings[ax.name]} "
            else:
                axis_mappings[ax.name] = next_index
                spec += f"{next_index} "
                next_index += 1

        array_specs.append(spec)

    # now compute the output axes:
    output_spec = " ".join(str(axis_mappings[ax.name]) for ax in output_axes)

    # get a name for jax so it's easier to interpret logs
    if axis is None:
        jax_str = f"contract {', '.join(axis_name(ax) for ax in all_axes)} -> <scalar>"
    else:
        axis = axis_spec_to_shape_dict(axis)
        jax_str = f"contract {', '.join(axis)} -> {', '.join(a.name for a in output_axes)}"

    with jax.named_scope(jax_str):
        output = _jittable_dg_einsum(
            ", ".join(array_specs) + "-> " + output_spec,
            *[a.array for a in arrays],
            precision=precision,
            preferred_element_type=preferred_element_type,
            _dot_general=dot_general,
        )

    out = NamedArray(output, output_axes)
    return haliax.auto_sharded(out)


def _ensure_no_mismatched_axes(*arrays: NamedArray):
    """Ensure that all the arrays have no axes with the same name but different sizes"""
    if len(arrays) <= 1:
        return

    known_sizes: dict[str, int] = {}
    for a in arrays:
        for ax in a.axes:
            if ax.name in known_sizes:
                if known_sizes[ax.name] != ax.size:
                    raise ValueError(f"Axis {ax.name} has multiple sizes: {known_sizes[ax.name]} and {ax.size}")
            else:
                known_sizes[ax.name] = ax.size



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


import pytest

from haliax._src.parsing import parse_einsum, parse_rearrangement


def _simplify_captures(expr):
    def simplify_capture(capture):
        if capture == Ellipsis:
            return Ellipsis
        elif (capture.binding == capture.axes[0] or capture.binding is None) and len(capture.axes) == 1:
            return capture.axes[0]
        elif capture.binding is None:
            return capture.axes
        else:
            return {capture.binding: capture.axes}

    return [simplify_capture(capture) for capture in expr.captures]


def test_parse_rearrangement_simple():
    lhs, rhs = parse_rearrangement("a b c d -> b c a d")
    assert lhs.is_ordered
    assert _simplify_captures(lhs) == ["a", "b", "c", "d"]
    assert rhs.is_ordered
    assert _simplify_captures(rhs) == ["b", "c", "a", "d"]

    lhs, rhs = parse_rearrangement("a ... c d -> b c a d")
    assert lhs.is_ordered
    assert _simplify_captures(lhs) == ["a", Ellipsis, "c", "d"]
    assert rhs.is_ordered
    assert _simplify_captures(rhs) == ["b", "c", "a", "d"]

    # longer identifiers
    lhs, rhs = parse_rearrangement("a_longer b123 c d -> b123 c a_longer d")
    assert lhs.is_ordered
    assert _simplify_captures(lhs) == ["a_longer", "b123", "c", "d"]
    assert rhs.is_ordered
    assert _simplify_captures(rhs) == ["b123", "c", "a_longer", "d"]


def test_parse_einsum_ordered():
    # We could support this syntax for dot with something like:
    #
    #     support normal einops syntax, including short name-capture: hax.dot("... c h w, h w d -> ...  c d", a, b)
    #     hax.dot("{h, w} -> ", a, b) means "contract h and w", analogous to hax.dot(a, b, axis=("h", "w"))
    #     hax.dot("{h, w} -> ... channel embed", a, b) means "contract h and w and ensure that the result ends with [channel, embed]" (by transposing/einsum)
    #     hax.dot(" -> batch channel embed", a, b) could mean "contract all but the named dims". Not entirely sure how I feel about that one, but used situationally it's probably ok

    lhses, rhs = parse_einsum("a b c d, b c e f -> a d e f")
    assert lhses is not None
    assert len(lhses) == 2
    assert all(lhs.is_ordered for lhs in lhses)
    lhs0_captures = _simplify_captures(lhses[0])
    lhs1_captures = _simplify_captures(lhses[1])
    assert lhs0_captures == ["a", "b", "c", "d"]
    assert lhs1_captures == ["b", "c", "e", "f"]
    assert rhs.is_ordered
    assert _simplify_captures(rhs) == ["a", "d", "e", "f"]

    lhses, rhs = parse_einsum("... c h w, h w d -> ...  c d")
    assert lhses is not None
    assert len(lhses) == 2
    assert all(lhs.is_ordered for lhs in lhses)
    lhs0_captures = _simplify_captures(lhses[0])
    lhs1_captures = _simplify_captures(lhses[1])
    assert lhs0_captures == [..., "c", "h", "w"]
    assert lhs1_captures == ["h", "w", "d"]
    assert rhs.is_ordered
    assert _simplify_captures(rhs) == [..., "c", "d"]

    lhses, rhs = parse_einsum("{...} -> batch channel embed")
    assert lhses is not None
    assert len(lhses) == 1
    assert not lhses[0].is_ordered
    assert _simplify_captures(lhses[0]) == [...]
    assert rhs.is_ordered
    assert _simplify_captures(rhs) == ["batch", "channel", "embed"]

    # just lhs
    lhses, rhs = parse_einsum("batch channel embed -> ")
    assert lhses is not None
    assert len(lhses) == 1
    assert lhses[0].is_ordered
    assert _simplify_captures(lhses[0]) == ["batch", "channel", "embed"]
    assert rhs.is_ordered
    assert _simplify_captures(rhs) == []

    # lhs x 2, empty rhs
    lhses, rhs = parse_einsum("batch channel embed, batch channel embed ->")
    assert lhses is not None
    assert len(lhses) == 2
    assert all(lhs.is_ordered for lhs in lhses)
    assert _simplify_captures(lhses[0]) == ["batch", "channel", "embed"]
    assert _simplify_captures(lhses[1]) == ["batch", "channel", "embed"]
    assert rhs.is_ordered
    assert _simplify_captures(rhs) == []


def test_parse_einsum_unordered():
    lhses, rhs = parse_einsum("{a, b} -> ")
    assert lhses is not None
    assert len(lhses) == 1
    assert not lhses[0].is_ordered
    assert _simplify_captures(lhses[0]) == ["a", "b"]
    assert rhs.is_ordered
    assert _simplify_captures(rhs) == []

    lhses, rhs = parse_einsum("{...} -> ")
    assert lhses is not None
    assert len(lhses) == 1
    assert not lhses[0].is_ordered
    assert _simplify_captures(lhses[0]) == [...]
    assert rhs.is_ordered
    assert _simplify_captures(rhs) == []

    lhses, rhs = parse_einsum("{h, w} -> ... channel embed")
    assert lhses is not None
    assert len(lhses) == 1
    assert not lhses[0].is_ordered
    assert _simplify_captures(lhses[0]) == ["h", "w"]
    assert rhs.is_ordered
    assert _simplify_captures(rhs) == [..., "channel", "embed"]


def test_parse_paren_groups():
    lhs, rhs = parse_rearrangement("a (b c) d -> b c a d")
    assert lhs.is_ordered
    assert _simplify_captures(lhs) == ["a", ("b", "c"), "d"]
    assert rhs.is_ordered
    assert _simplify_captures(rhs) == ["b", "c", "a", "d"]

    lhs, rhs = parse_rearrangement("a (b: c) (d: e f) -> b c a d")
    assert lhs.is_ordered
    assert _simplify_captures(lhs) == ["a", {"b": ("c",)}, {"d": ("e", "f")}]


def test_parse_unordered():
    lhs, rhs = parse_rearrangement("{a b c d} -> {b c a d}")
    assert not lhs.is_ordered
    assert _simplify_captures(lhs) == ["a", "b", "c", "d"]
    assert not rhs.is_ordered
    assert _simplify_captures(rhs) == ["b", "c", "a", "d"]

    lhs, rhs = parse_rearrangement("{(c: a b) d e} -> (q: a d e) b")
    assert not lhs.is_ordered
    assert _simplify_captures(lhs) == [{"c": ("a", "b")}, "d", "e"]
    assert rhs.is_ordered
    assert _simplify_captures(rhs) == [{"q": ("a", "d", "e")}, "b"]


def test_parse_quoted_identifiers():
    lhs, rhs = parse_rearrangement("a \"b c\" d -> 'b c' a d")
    assert lhs.is_ordered
    assert _simplify_captures(lhs) == ["a", "b c", "d"]
    assert rhs.is_ordered
    assert _simplify_captures(rhs) == ["b c", "a", "d"]

    lhs, rhs = parse_rearrangement("{a \"b c\" (d: 'hello')} -> b c a d")
    assert not lhs.is_ordered
    assert _simplify_captures(lhs) == ["a", "b c", {"d": ("hello",)}]
    assert rhs.is_ordered
    assert _simplify_captures(rhs) == ["b", "c", "a", "d"]


def test_parse_errors():
    with pytest.raises(ValueError, match="Unexpected end of string"):
        parse_rearrangement("a b")

    with pytest.raises(ValueError, match="Expected }"):
        parse_rearrangement("{ a -> c")

    with pytest.raises(ValueError, match="Unexpected }"):
        parse_rearrangement("a } -> c")

    with pytest.raises(ValueError, match="Unexpected }"):
        parse_rearrangement("(a: b } -> c")

    with pytest.raises(ValueError, match="Unexpected character"):
        parse_rearrangement("(a b ! -> c d e")

    with pytest.raises(ValueError, match="Identifier expected"):
        parse_rearrangement("a b ! -> c d e")
